"""Parse castep .cell and .param files."""

from __future__ import annotations

import re
import typing
from collections import Counter, defaultdict
from collections.abc import Callable, Sequence
from functools import partial
from typing import Any, Literal, TextIO, TypedDict, Union

import castep_outputs.utilities.castep_res as REs

from ..utilities.constants import SHELLS
from ..utilities.datatypes import (
    AtomIndex,
    MaybeSequence,
    PSPotStrInfo,
    ThreeByThreeMatrix,
    ThreeVector,
)
from ..utilities.filewrapper import Block
from ..utilities.utility import (
    atreg_to_index,
    determine_type,
    file_or_path,
    fix_data_types,
    log_factory,
    normalise_key,
    strip_comments,
    to_type,
)


class PositionsInfo(TypedDict, total=False):
    """Information from positions block."""

    #: Units.
    units: str
    #: Atomic position.
    pos: ThreeVector
    #: Ion spin.
    spin: float
    #: Mixture index.
    mix_index: int
    #: Mixture weight.
    weight: float


class NonLinearConstraint(TypedDict):
    """Non-linear constraint information."""

    #: Type of non-linear constraint.
    key: Literal["distance", "bend", "torsion"]
    #: Atoms and constraint definition.
    atoms: dict[AtomIndex, ThreeVector]


class XCDefParams(TypedDict, total=False):
    """XC definition params."""

    nlxc_screening_length: float
    nlxc_screening_function: str
    nlxc_ppd_int: bool
    nlxc_divergence_corr: bool


class XCDef(TypedDict):
    """Information from XC definitions block."""

    params: XCDefParams
    xc: dict[str, float]


DevelElem = MaybeSequence[Union[str, float, dict[str, Union[str, float]]]]
DevelBlock = dict[str, Union[DevelElem, dict[str, DevelElem]]]
HubbardU = dict[Union[str, AtomIndex], Union[str, dict[str, float]]]
CellParamData = dict[
    str, Union[str, float, tuple[float, str], dict[str, Any], HubbardU, DevelBlock, XCDef],
]
GeneralBlock = dict[
    str,
    Union[
        list[Union[str, float]],
        dict[str, MaybeSequence[float]],
    ],
]


def _get_block_units(block: Block, default: str) -> str:
    """Get units from a block.

    Consume the units line if present else rewind.

    Parameters
    ----------
    block
        Block to search for units.
    default
        Unit to use if not present.

    Returns
    -------
    str
        Units.
    """
    test = next(block)
    if match := re.match(r"^\s*(\S+)\s*$", test):
        units = match[1]
    else:
        block.rewind()
        units = default

    return units


@file_or_path(mode="r")
def parse_cell_param_file(cell_param_file: TextIO) -> list[CellParamData]:
    """
    Parse castep .cell and param files.

    Parameters
    ----------
    cell_param_file
        Open handle to file to parse.

    Returns
    -------
    list[dict[str, str | dict[str, str]]]
        Parsed info.
    """
    logger = log_factory(cell_param_file)
    curr: CellParamData = {}

    for line in cell_param_file:
        # Strip comments
        line = re.split(r"[#!]", line)[0]

        if block := Block.from_re(
            line,
            cell_param_file,
            re.compile(r"^\s*%block ", re.IGNORECASE),
            re.compile(r"^\s*%endblock", re.IGNORECASE),
        ):
            block_title = next(block).split()[1].lower()

            block = strip_comments(block, remove_inline=True)
            logger("Found block %s", block_title)

            curr[block_title] = _PARSERS.get(block_title, _parse_general)(block)

        elif match := REs.PARAM_VALUE_RE.match(line):
            key, val, unit = match.group("key", "val", "unit")
            key = key.strip().lower()
            logger("Found param %s", key)

            if " " in val.strip():
                val = to_type(val.split(), determine_type(val))
            else:
                val = to_type(val, determine_type(val))
            if unit:
                unit = unit.strip()
                curr[key] = (val, unit)
            else:
                curr[key] = val

    return [curr]


parse_cell_file = parse_cell_param_file
parse_param_file = parse_cell_param_file


def _parse_pspot_string(string: str, *, debug: bool = False) -> PSPotStrInfo:
    if not (match := REs.PSPOT_RE.search(string)):
        raise ValueError(f"Attempt to parse {string} as PSPot failed")

    pspot = match.groupdict()
    projectors = []

    for proj in pspot["proj"].split(":"):
        if match := REs.PSPOT_PROJ_RE.match(proj):
            pdict = dict(zip(REs.PSPOT_PROJ_GROUPS, match.groups()))
        else:
            raise ValueError("Invalid PSPot string")

        pdict["shell"] = SHELLS[int(pdict["shell"])]

        if not pdict["type"]:
            pdict["type"] = None

        for prop in ("beta_delta", "de"):
            if not pdict[prop]:
                del pdict[prop]

        fix_data_types(pdict, {"orbital": int, "beta_delta": float, "de": float})
        projectors.append(pdict)

    for prop in ("shell_swp", "shell_swp_end", "opt"):
        if pspot[prop]:
            pspot[prop] = pspot[prop].split(",")

    pspot["projectors"] = tuple(projectors)
    pspot["string"] = string
    pspot["print"] = bool(pspot["print"])

    if not debug:
        for prop in (
            "shell_swp",
            "shell_swp_end",
            "local_energy",
            "poly_fit",
            "beta_radius",
            "r_inner",
            "debug",
        ):
            if pspot[prop] is None:
                del pspot[prop]

    fix_data_types(
        pspot,
        {
            "beta_radius": float,
            "r_inner": float,
            "core_radius": float,
            "coarse": float,
            "medium": float,
            "fine": float,
            "local_channel": int,
        },
    )

    return pspot


def _parse_devel_code_block(in_block: Block) -> DevelBlock:
    """
    Parse devel_code block to dict.

    Parameters
    ----------
    in_block
        Input block to parse.

    Returns
    -------
    dict[str, str | float]
        Parsed info.
    """
    main_block = " ".join(map(str.strip, in_block))

    matches = re.finditer(REs.DEVEL_CODE_BLOCK_GENERIC_RE, main_block, re.IGNORECASE | re.MULTILINE)
    devel_code_parsed: DevelBlock = {}
    par: re.Match[str] | str

    # Get groups
    for blk in matches:
        block_title = normalise_key(blk.group(1))
        block_str = blk.group(0).split(":")[1]
        block: DevelBlock = {}
        for par in re.finditer(REs.DEVEL_CODE_VAL_RE, block_str, re.IGNORECASE | re.MULTILINE):
            key, val = re.split(r"[:=]", par.group(0))
            key = normalise_key(key)
            typ = determine_type(val)
            block[key] = to_type(val, typ)
            block_str = block_str.replace(par.group(0), "")

        if block_str.strip():
            block["data"] = []
            for par in block_str.split():
                assert isinstance(block["data"], list)
                block["data"].append(to_type(par, determine_type(par)))

        if block_title in devel_code_parsed:
            devel_code_parsed[block_title].update(block)  # type: ignore
        else:
            devel_code_parsed[block_title] = block  # type: ignore

        # Remove matched to get remainder
        main_block = main_block.replace(blk.group(0), "")

    # Catch remainder
    for par in re.finditer(REs.DEVEL_CODE_VAL_RE, main_block):
        key, val = re.split(r"[:=]", par.group(0))
        key = normalise_key(key)
        typ = determine_type(val)

        if key in devel_code_parsed:  # Var has same name as block
            key = f"_{key}"

        devel_code_parsed[key] = to_type(val, typ)

    return devel_code_parsed


def _parse_ionic_constraints(block: Block) -> dict[AtomIndex, Sequence[ThreeVector]]:
    """
    Parse ionic constraints to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[AtomIndex, ThreeVector]
        Parsed info.
    """
    accum = defaultdict(list)

    for line in block:
        if match := REs.IONIC_CONSTRAINTS_RE.match(line):
            ind = atreg_to_index(match)
            elem: ThreeVector = to_type(match["val"].split(), float)
            accum[ind].append(elem)

    return dict(accum)


def _parse_nonlinear_constraints(block: Block) -> list[NonLinearConstraint]:
    """
    Parse nonlinear constraints to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[AtomIndex, ThreeVector]
        Parsed info.
    """
    accum = []

    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        key, line = line.split(maxsplit=1)
        cons: NonLinearConstraint = {
            "key": key,
            "atoms": {
                atreg_to_index(match): to_type(match.group("x", "y", "z"), int)
                for match in REs.ATOMIC_DATA_3VEC.finditer(line)
            },
        }
        accum.append(cons)

    return accum


def _parse_positions(block: Block, *, absolute: bool = False) -> dict[AtomIndex, PositionsInfo]:
    """
    Parse positions to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[AtomIndex, ThreeVector]
        Parsed info.
    """
    accum = {}
    count: Counter = Counter()

    if absolute:
        units = _get_block_units(block, "Ang")

    for line in block:
        if match := REs.POSITIONS_LINE_RE.match(line):
            spec: str = match["spec"]
            count[spec] += 1
            ind = (spec, count[spec])
            pos = to_type(match["pos"].split(), float)

            info: PositionsInfo = {"pos": pos}
            if absolute:
                info["units"] = units

            if match := REs.POSITIONS_SPIN_RE.search(line):
                info["spin"] = to_type(match["spin"], float)

            if match := REs.POSITIONS_MIXTURE_RE.search(line):
                info["mix_index"] = to_type(match["mix"], int)
                info["weight"] = to_type(match["ratio"], float)

            accum[ind] = info

    return accum


def _parse_hubbard_u(block: Block) -> HubbardU:
    """
    Parse Hubbard U block to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[str | AtomIndex, dict[str, float]]
        Parsed info.
    """
    accum: HubbardU = {"units": _get_block_units(block, "eV")}

    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        if match := re.match(rf"^\s*({REs.ATOM_NAME_RE})\s*(\d+)?", line):
            spec = (match[1], int(match[2])) if match.lastindex == 2 else match[1]
            line = line.replace(match[0], "")
            accum[spec] = {
                key: to_type(val, determine_type(val))
                for key, val in REs.COLON_SEP_RE.findall(line)
            }

    return accum


def _parse_sedc(block: Block) -> dict[str, dict[str, float]]:
    """
    Parse Semi-empirical dispersion correction block to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[str, dict[str, float]]
        Parsed info.
    """
    accum = {}

    test = next(block)
    if len(units := test.split()) == 2:
        units = tuple(units)
    else:
        block.rewind()
        units = ("eV", "Ang")

    for line in block:
        if re.match(r"^\s*%endblock", line, re.IGNORECASE):
            continue

        spec, line = line.strip().split(maxsplit=1)
        accum[spec] = {
            key: to_type(val, determine_type(val)) for key, val in REs.COLON_SEP_RE.findall(line)
        }

    return accum


def _parse_symops(block: Block) -> list[dict[str, ThreeByThreeMatrix | ThreeVector]]:
    """
    Parse symmetry operations block to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    list[dict[str, ThreeByThreeMatrix | ThreeVector]]
        Parsed info.
    """
    tmp = [to_type(numbers, float) for line in block if (numbers := REs.FLOAT_RAT_RE.findall(line))]

    return [
        {
            "r": tmp[i : i + 3],  # type: ignore
            "t": tmp[i + 3],
        }
        for i in range(0, len(tmp), 4)
    ]


def _parse_xc_definition(block: Block) -> dict[str, float]:
    """
    Parse xc_definition block to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[str, float]
        Parsed info.
    """
    block_data = {"xc": {}, "params": {}}
    block.remove_bounds(fore=0, back=1)

    for line in strip_comments(block):
        key, val = line.split(maxsplit=1)
        key = normalise_key(key)

        if key == "nlxc_screening_length":
            block_data["params"][key] = float(val)
        elif key == "nlxc_screening_function":
            block_data["params"][key] = val
        elif key in {"nlxc_ppd_int", "nlxc_divergence_corr"}:
            block_data["params"][key] = val.upper() == "ON"
        else:
            block_data["xc"][key] = float(val)

    return block_data


def _parse_species_pot(block: Block) -> dict[str, str | PSPotStrInfo]:
    """Parse species pot block.

    Parameters
    ----------
    block : Block
        Input block to parse.

    Returns
    -------
    dict[str, str | PSPotStrInfo]
        Processed block if PSPot string, otherwise bare string.
    """
    block_data = {}
    block.remove_bounds(fore=0, back=1)

    for line in block:
        spec, pot = line.split(maxsplit=1)
        if REs.PSPOT_RE.search(pot):  # We have pspot definition
            pot = _parse_pspot_string(pot)
        block_data[spec] = pot

    return block_data


def _parse_general(block: Block) -> GeneralBlock:
    """
    Parse general block to dict.

    Parameters
    ----------
    block
        Input block to parse.

    Returns
    -------
    dict[str, str | float]
        Parsed info.
    """
    block_data: GeneralBlock = {"data": []}

    for line in block:
        line = re.split(r"[#!]", line)[0]

        if REs.SPEC_PROP_RE.match(line):
            if isinstance(block_data["data"], list):
                block_data["data"] = {}
                typing.cast(dict[str, MaybeSequence[Union[float, str]]], block_data["data"])

            spec, val = line.strip().split(maxsplit=1)
            val = to_type(val, determine_type(val))

            typing.cast(MaybeSequence[Union[float, str]], val)

            block_data["data"][spec] = val

        elif numbers := REs.FLOAT_RAT_RE.findall(line):
            assert isinstance(block_data["data"], list)
            block_data["data"].append(to_type(numbers, float))

        elif re.match(r"^\s*%endblock", line, re.IGNORECASE):
            pass

        else:
            block_data["units"] = line.strip()

    return block_data


_parse_positions_abs = partial(_parse_positions, absolute=True)
_parse_positions_frac = partial(_parse_positions, absolute=False)

#: Cell/Param subparsers.
_PARSERS: dict[str, Callable] = {
    "devel_code": _parse_devel_code_block,
    "ionic_constraints": _parse_ionic_constraints,
    "nonlinear_constraints": _parse_nonlinear_constraints,
    "positions_abs": _parse_positions_abs,
    "positions_frac": _parse_positions_frac,
    "positions_abs_intermediate": _parse_positions,
    "positions_frac_intermediate": _parse_positions,
    "positions_abs_product": _parse_positions,
    "positions_frac_product": _parse_positions,
    "species_pot": _parse_species_pot,
    "sedc_custom_params": _parse_sedc,
    "hubbard_u": _parse_hubbard_u,
    "symmetry_ops": _parse_symops,
    "xc_definition": _parse_xc_definition,
}



"""Parser for cst_esp files."""
from __future__ import annotations

from typing import BinaryIO, TypedDict

from ..utilities.utility import file_or_path, to_type
from .fortran_bin_parser import binary_file_reader


class ESPData(TypedDict):
    """Data from electrostatic potential."""

    #: Number of spins in run.
    n_spins: int
    #: Grid size sampled at.
    grid: int
    #: ESP Data.
    esp: tuple[tuple[tuple[complex, ...], ...], ...]
    #: MGGA
    mgga: tuple[tuple[tuple[complex, ...], ...], ...]


@file_or_path(mode="rb")
def parse_cst_esp_file(cst_esp_file: BinaryIO) -> ESPData:
    """Parse castep `cst_esp` files.

    Parameters
    ----------
    cst_esp_file : BinaryIO
        File to parse.

    Returns
    -------
    ESPData
        Parsed data.
    """
    dtypes = {"n_spins": int, "grid": int}

    accum = {"esp": []}

    reader = binary_file_reader(cst_esp_file)
    for (key, typ), datum in zip(dtypes.items(), reader):
        accum[key] = to_type(datum, typ)

    prev_nx = None
    curr = []
    for datum in reader:
        nx, _ny = to_type(datum[:8], int)
        if prev_nx != nx and curr:
            accum["esp"].append(curr)
            curr = []
        curr.append(to_type(datum[8:], complex))
        prev_nx = nx

    accum["esp"].append(curr)

    size = accum["grid"][0]
    if len(accum["esp"]) > size:  # Have MGGA pot
        accum["esp"], accum["mgga"] = accum["esp"][:size], accum["esp"][size:]

    return accum



"""General parser for the Fortran Unformatted file format."""
from collections.abc import Generator
from os import SEEK_CUR
from typing import BinaryIO

FortranBinaryReader = Generator[bytes, int, None]


def binary_file_reader(file: BinaryIO) -> FortranBinaryReader:
    """Yield the elements of a Fortran unformatted file.

    Parameters
    ----------
    file : BinaryIO
        Open file to get binary data from.

    Yields
    ------
    bytes
        Binary data record from Fortran file.

    Receives
    --------
    int | bool
        Skip/rewind amount (``True`` == ``-1``).

    Notes
    -----
    Each "record" is:

    ``(pre_nbytes: 4; data: nbytes; post_nbytes: 4)``

    Where ``pre_nbytes == post_nbytes`` (this is used in Fortran for rewinding).

    So when we rewind, we rewind the current size + ``post_nbytes``
    (current) + ``pre_nbytes`` (current) + ``post_nbytes`` (previous) [cursor
    now before post_nbytes (previous)] which is then read putting the
    cursor after post_nbytes (previous).

    When we do the final rewind, we put the cursor before ``pre_nbytes``
    ready to read the record.
    """
    while bin_size := file.read(4):
        size = int.from_bytes(bin_size, "big")
        data = file.read(size)
        skip = yield data
        file.read(4)
        if skip:  # NB. Send proceeds to yield.
            # `True` implies rewind 1
            if skip < 0 or skip is True:
                for _ in range(abs(skip) - 1):
                    # Rewind to record size before last read
                    file.seek(-size - 12, SEEK_CUR)
                    size = int.from_bytes(file.read(4), "big")

                # Rewind one extra (which will be yielded)
                file.seek(-size - 8, SEEK_CUR)
            else:
                for _ in range(skip):
                    size = int.from_bytes(file.read(4), "big")
                    file.seek(size + 4, SEEK_CUR)



"""Test get_generated_files."""
import pytest

from castep_outputs.tools.get_generated_files import get_generated_files

DUMMY_CELL = {
    "positions_frac": {
        ("Si", 1): {},
        ("N", 1): {},
    },
}


@pytest.mark.parametrize("seedname, expected", [
    ("test", [
        "test.bands",
        "test.bib",
        "test.castep",
        "test.castep_bin",
        "test.check",
        "test.cst_esp",
    ]),
])
def test_cli_like(seedname: str, expected: list[str]) -> None:
    """Test via the CLI like interface parsing files on-the-fly."""
    assert expected == get_generated_files(seedname)


@pytest.mark.parametrize("cell, param, expected", [
    (DUMMY_CELL,
     {
         "task": "moleculardynamics",
         "metals_method": "dm",
         "opt_strategy": "memory",
         "xc_functional": "LDA",
         "md_ensemble": "NVE",
         "md_extrap": "mixed",
     },
     {
         "seedname.*.drho2m",
         "seedname.bib",
         "seedname.cst_esp",
         "seedname.*.drhom",
         "seedname.castep",
         "seedname.md",
         "seedname.*.wf2m",
         "seedname.castep_bin",
         "seedname.*.wfm",
         "seedname.bands",
         "seedname.check",
     }),

    (DUMMY_CELL | {
        "species_pot": {
            "Si": {
                "proj": "30N1.9:31N2.1:32N2.1",
                "opt": [
                    "tm",
                    "nonlcc",
                ],
                "print": True,
                "projectors": [
                    {
                        "orbital": 3,
                        "shell": "s",
                        "type": "N",
                        "de": 1.9,
                    },
                    {
                        "orbital": 3,
                        "shell": "p",
                        "type": "N",
                        "de": 2.1,
                    },
                    {
                        "orbital": 3,
                        "shell": "d",
                        "type": "N",
                        "de": 2.1,
                    },
                ],
                "string": "2|1.9|2.1|1.3|3.675|5.512|7.35|30N1.9:31N2.1:32N2.1(tm,nonlcc)[]",
            },
        },
    },
     {
         "task": "singlepoint",
         "metals_method": "dm",
         "opt_strategy": "memory",
         "xc_definition": {
             "xc": {
                 "LIBXC_GGA_XC_B97_D": 1.0,
             },
         },
         "write_otfg": "T",
     },
     {
         "seedname.castep",
         "seedname.cst_esp",
         "Si_EXT_PBE_OTF.gamma",
         "Si_EXT_PBE_OTF.pwave",
         "Si_EXT_PBE_OTF.econv",
         "Si_EXT_PBE_OTF.usp",
         "seedname.bands",
         "Si_EXT_PBE_OTF.beta",
         "seedname.bib",
         "seedname.castep_bin",
         "seedname.check",
     }),


    (DUMMY_CELL,
     {
         "task": "energy",
         "write_checkpoint": "none",
         "write_bib": False,
         "devel_code": {
             "pp": {
                 "buck": True,
                 "pot_print": True,
                 "fd_check": True,
             },
             "_pp": True,
         },
     },
     {
         "seedname.castep",
         "seedname.fd",
         "seedname_N_N.pot",
         "seedname_Si_Si.pot",
         "seedname_Si_N.pot",
     },
    ),

    (DUMMY_CELL,
     {
         "task": "spectral",
         "xc_functional": "LDA",
         "max_scf_cycles": 30,
         "spectral_task": "bandstructure",
         "opt_strategy": "speed",
         "popn_calculate": False,
         "spin_treatment": "VECTOR",
         "spin_orbit_coupling": True,
         "popn_write": "minimal",
     },
     {
         "seedname-out.cell",
         "seedname.bands",
         "seedname.castep",
         "seedname.cst_esp",
         "seedname.*.spec",
         "seedname.bib",
         "seedname.castep_bin",
         "seedname.check",
     },
    ),

    (DUMMY_CELL,
     {
         "task": "ElectronicSpectroscopy",
         "spectral_task": "BandStructure",
         "xc_functional": "PBE",
         "spin_polarized": False,
         "spin_orbit_coupling": False,
         "opt_strategy": "Default",
         "metals_method": "dm",
         "mixing_scheme": "Pulay",
         "calculate_stress": False,
         "calculate_elf": False,
         "popn_calculate": False,
         "calculate_hirshfeld": False,
         "calculate_densdiff": False,
         "pdos_calculate_weights": True,
         "spectral_write_eigenvalues": True,
     },
     {
         "seedname-out.cell",
         "seedname.*.spec",
         "seedname.bib",
         "seedname.castep_bin",
         "seedname.check",
         "seedname.bands",
         "seedname.castep",
         "seedname.cst_esp",
         "seedname.pdos_bin",
     },
    ),

  (
      {
          "positions_frac": {
              "Si_1": {},
              "Si_2": {},
          },
          "species_pot": {
              "Si": {
                  "proj": "30N1.9:31N2.1:32N2.1",
                  "opt": [
                      "tm",
                      "nonlcc",
                  ],
                  "print": True,
                  "projectors": [
                      {
                          "orbital": 3,
                          "shell": "s",
                          "type": "N",
                          "de": 1.9,
                      },
                      {
                          "orbital": 3,
                          "shell": "p",
                          "type": "N",
                          "de": 2.1,
                      },
                      {
                          "orbital": 3,
                          "shell": "d",
                          "type": "N",
                          "de": 2.1,
                      },
                  ],
                  "string": "2|1.9|2.1|1.3|3.675|5.512|7.35|30N1.9:31N2.1:32N2.1(tm,nonlcc)[]",
              },
          },
      },
      {
          "opt_strategy": "speed",
          "calculate_stress": True,
          "grid_scale": 1.75,
      },
      {
          "seedname.bands",
          "seedname.castep",
          "seedname.cst_esp",
          "seedname.bib",
          "seedname.castep_bin",
          "seedname.check",
          "Si_EXT_LDA_OTF.beta",
          "Si_EXT_LDA_OTF.econv",
          "Si_EXT_LDA_OTF.gamma",
          "Si_EXT_LDA_OTF.pwave",
          "Si_EXT_LDA_OTF.usp",
      },
  ),

], ids=["MD/Si8-md-extrapm", "LIBXC test", "Pair_pot/PP_BUCK",
        "Excitations/GaAs", "Excitations/BN_DOS", "OTF-Gen/Si2-ps-tm"])
def test_expected_files(cell, param, expected) -> None:
    assert expected == set(get_generated_files(param, cell))



"""
Extract results from .castep file for comparison and further processing.

Notes
-----
Port of extract_results.pl
"""
from __future__ import annotations

import itertools
import re
from collections import defaultdict
from collections.abc import Callable
from enum import Flag, auto
from typing import Any, TextIO, Union, cast

from ..utilities import castep_res as REs
from ..utilities.castep_res import gen_table_re, get_numbers, labelled_floats
from ..utilities.constants import SHELLS
from ..utilities.datatypes import (
    AtomIndex,
    AtomPropBlock,
    BandStructure,
    BondData,
    CellInfo,
    CharTable,
    ConstraintsReport,
    DelocActiveSpace,
    DelocInternalsTable,
    DipoleTable,
    ElasticProperties,
    FinalConfig,
    GeomTable,
    InitialSpin,
    InternalConstraints,
    KPointsList,
    KPointsSpec,
    MDInfo,
    MemoryEst,
    MullikenInfo,
    Occupancies,
    PhononSymmetryReport,
    PSPotEnergy,
    PSPotReport,
    QData,
    RamanReport,
    SCFDebugInfo,
    SCFReport,
    SixVector,
    SymmetryReport,
    TDDFTData,
    Thermodynamics,
    ThreeByThreeMatrix,
    ThreeVector,
    WvfnLineMin,
)
from ..utilities.filewrapper import Block, FileWrapper
from ..utilities.utility import (
    Logger,
    add_aliases,
    atreg_to_index,
    determine_type,
    file_or_path,
    fix_data_types,
    get_only,
    log_factory,
    normalise_key,
    normalise_string,
    parse_int_or_float,
    stack_dict,
    to_type,
)
from .bands_file_parser import parse_bands_file
from .cell_param_file_parser import _parse_devel_code_block, _parse_pspot_string
from .efield_file_parser import parse_efield_file
from .elastic_file_parser import parse_elastic_file
from .hug_file_parser import parse_hug_file
from .parse_fmt_files import (
    parse_chdiff_fmt_file,
    parse_den_fmt_file,
    parse_elf_fmt_file,
    parse_pot_fmt_file,
)
from .phonon_dos_file_parser import parse_phonon_dos_file
from .xrd_sf_file_parser import parse_xrd_sf_file

# Reduced set of parsers needed for .castep test extras
PARSERS: dict[str, Callable] = {
    "bands": parse_bands_file,
    "hug": parse_hug_file,
    "phonon_dos": parse_phonon_dos_file,
    "efield": parse_efield_file,
    "xrd_sf": parse_xrd_sf_file,
    "elf_fmt": parse_elf_fmt_file,
    "chdiff_fmt": parse_chdiff_fmt_file,
    "pot_fmt": parse_pot_fmt_file,
    "den_fmt": parse_den_fmt_file,
    "elastic": parse_elastic_file,
}


class Filters(Flag):
    """Enum of possible filters for CASTEP file parsing."""

    BS = auto()
    CELL = auto()
    CHEM_SHIELDING = auto()
    DIPOLE = auto()
    ELASTIC = auto()
    ELF = auto()
    FINAL_CONFIG = auto()
    FORCE = auto()
    GEOM_OPT = auto()
    MD = auto()
    MD_SUMMARY = auto()
    OPTICS = auto()
    PARAMETERS = auto()
    PHONON = auto()
    POLARISATION = auto()
    POPN_ANALYSIS = auto()
    POSITION = auto()
    PSPOT = auto()
    SCF = auto()
    SOLVATION = auto()
    SPECIES_PROPS = auto()
    SPIN = auto()
    STRESS = auto()
    SYMMETRIES = auto()
    SYS_INFO = auto()
    TDDFT = auto()
    TEST_EXTRA_DATA = auto()
    THERMODYNAMICS = auto()
    TSS = auto()

    # Preset sets
    NONE = 0
    LOW = (BS | CELL | CHEM_SHIELDING | DIPOLE |
           ELASTIC | ELF | FINAL_CONFIG | FORCE |
           MD_SUMMARY | OPTICS | POPN_ANALYSIS |
           POSITION | POLARISATION | SOLVATION | SPECIES_PROPS | SPIN |
           STRESS | TDDFT | THERMODYNAMICS | TSS)

    MEDIUM = LOW | PARAMETERS | GEOM_OPT | MD | PHONON

    HIGH = MEDIUM | PSPOT | SYMMETRIES | SYS_INFO

    FULL = HIGH | SCF
    ALL = FULL

    TESTING = (BS | CELL | CHEM_SHIELDING | DIPOLE |
               ELASTIC | ELF | FORCE |
               GEOM_OPT | MD | OPTICS |
               PHONON | POPN_ANALYSIS | POSITION |
               PSPOT | SOLVATION | SPECIES_PROPS |
               STRESS | TDDFT | TEST_EXTRA_DATA |
               THERMODYNAMICS | TSS)


@file_or_path(mode="r")
def parse_castep_file(castep_file_in: TextIO,
                      filters: Filters = Filters.HIGH) -> list[dict[str, Any]]:
    """
    Parse castep file into lists of dicts ready to JSONise.

    Parameters
    ----------
    castep_file_in
        File to parse.
    filters
        Parameters to parse.

    Returns
    -------
    list[dict[str, Any]]
        Parsed data.

    Raises
    ------
    ValueError
        On invalid top-level blocks.
    """
    # pylint: disable=redefined-outer-name

    runs: list[dict[str, Any]] = []
    curr_run: dict[str, Any] = defaultdict(list)

    to_parse = filters

    if not isinstance(castep_file_in, (FileWrapper, Block)):
        castep_file = FileWrapper(castep_file_in)
    else:
        castep_file = castep_file_in

    logger = log_factory(castep_file)

    for line in castep_file:
        logger("%s", line, level="debug")
        # Build Info
        if block := Block.from_re(line, castep_file,
                                  r"^\s*Compiled for",
                                  REs.EMPTY):

            if curr_run:
                runs.append(curr_run)
            logger("Found run %s", len(runs) + 1)
            curr_run = defaultdict(list)

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found build info")
            curr_run["build_info"] = _process_buildinfo(block)

        elif "Run started" in line:

            if Filters.SYS_INFO not in to_parse:
                continue

            curr_run["time_started"] = normalise_string(line.split(":", 1)[1])

        # Finalisation
        elif block := Block.from_re(line, castep_file, "Initialisation time", REs.EMPTY,
                                    eof_possible=True):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found finalisation")
            curr_run.update(_process_finalisation(block))

        elif line.startswith("Overall parallel efficiency rating"):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found parallel efficiency")

            curr_run["parallel_efficiency"] = float(get_numbers(line)[0])

        # Continuation
        elif line.startswith("Reading continuation data"):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found continuation block")

            curr_run["continuation"] = line.split()[-1]

        # Warnings
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("", r"\?+"),
                                    gen_table_re("", r"\?+")):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found warning")

            block.remove_bounds(1, 1)
            curr_run["warning"].append(" ".join(x.strip() for x in block))

        elif match := re.match(r"(?:\s*[^:]+:)?(\s*)warning", line, re.IGNORECASE):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found warning")

            warn = line.strip()

            for tst, line in enumerate(castep_file):
                if not line.strip() or not re.match(match.group(1) + r"\s+", line):
                    if tst:
                        castep_file.rewind()
                    break
                warn += " " + line.strip()

            curr_run["warning"].append(warn)

        # Memory estimate
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re(r"MEMORY AND SCRATCH[\w\s]+", "[+-]+"),
                                    gen_table_re("", "[+-]+")):

            if Filters.SYS_INFO not in to_parse:
                continue

            logger("Found memory estimate")

            curr_run["memory_estimate"].append(_process_memory_est(block))

        # Title
        elif re.match(gen_table_re("Title", r"\*+"), line):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found title")

            curr_run["title"] = next(castep_file).strip()

        # Parameters
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("[^*]+ Parameters", r"\*+"),
                                    gen_table_re("", r"\*+")):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found options")

            curr_run["options"] = _process_params(block)

        # Quantisation axis
        elif "Quantisation axis" in line:

            if Filters.SPECIES_PROPS not in to_parse:
                continue

            logger("Found Quantisation axis")

            curr_run["quantisation_axis"] = to_type(get_numbers(line), float)

        # Pseudo-atomic energy
        elif block := Block.from_re(line, castep_file, REs.PS_SHELL_RE, REs.EMPTY, n_end=2):

            if Filters.SPECIES_PROPS not in to_parse:
                continue

            logger("Found pseudo-atomic energy")

            key, val = _process_ps_energy(block)

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            curr_run["species_properties"][key].update(val)

        # Mass
        elif block := Block.from_re(line, castep_file, r"Mass of species in AMU", REs.EMPTY):

            if Filters.SPECIES_PROPS not in to_parse:
                continue

            logger("Found mass")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val in _process_spec_prop(block):
                curr_run["species_properties"][key]["mass"] = float(val)

        # Electric Quadrupole Moment
        elif block := Block.from_re(line, castep_file,
                                    r"Electric Quadrupole Moment",
                                    rf"({REs.EMPTY}|^\s*x+$)"):
            if Filters.SPECIES_PROPS not in to_parse:
                continue

            logger("Found electric quadrupole moment")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val, *_ in _process_spec_prop(block):
                curr_run["species_properties"][key]["electric_quadrupole_moment"] = float(val)

        # Pseudopots
        elif block := Block.from_re(line, castep_file,
                                    r"Files used for pseudopotentials", REs.EMPTY):

            if Filters.SPECIES_PROPS not in to_parse:
                continue

            logger("Found pseudopotentials")

            if "species_properties" not in curr_run:
                curr_run["species_properties"] = defaultdict(dict)

            for key, val in _process_spec_prop(block):
                if Filters.PSPOT in to_parse and "|" in val:
                    val = _parse_pspot_string(val)

                curr_run["species_properties"][key]["pseudopot"] = val

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Pseudopotential Report[^|]+", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.PSPOT not in to_parse:
                continue

            logger("Found pseudopotential report")

            curr_run["pspot_detail"].append(_process_pspot_report(block))

        elif match := re.match(r"\s*(?P<type>AE|PS) eigenvalue nl (?P<nl>\d+) =" +
                               labelled_floats(("eigenvalue",)), line):

            if Filters.PSPOT not in to_parse:
                continue

            logger("Found PSPot debug for %s at %s", match["type"], match["nl"])

            val = match.groupdict()
            fix_data_types(val, {"nl": int, "eigenvalue": float})

            curr_run["pspot_debug"].append(val)

        # Pair Params
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("PairParams", r"\*+", pre=r"\w*"),
                                    REs.EMPTY):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found pair params")

            curr_run["pair_params"].append(_process_pair_params(block))

        # DFTD
        elif block := Block.from_re(line, castep_file,
                                    "DFT-D parameters",
                                    r"^\s*$", n_end=3):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found DFTD block")

            curr_run["dftd"] = _process_dftd(block)

        # SCF
        elif block := Block.from_re(line, castep_file, "SCF loop", "^-+ <-- SCF", n_end=2):

            if Filters.SCF not in to_parse:
                continue

            logger("Found SCF")

            curr_run["scf"].append(_process_scf(block))

        # SCF Line min
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("WAVEFUNCTION LINE MINIMISATION", "[+-]+",
                                                 post="<- line"),
                                    gen_table_re("", "[+-]+", post="<- line"), n_end=2):

            if Filters.SCF not in to_parse:
                continue

            logger("Found wvfn line min")

            curr_run["wvfn_line_min"].append(_process_wvfn_line_min(block))

        # SCF Occupancy
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Occupancy", r"\|",
                                                 post="<- occ", whole_line=False),
                                    r"Have a nice day\."):

            if Filters.SCF not in to_parse:
                continue

            logger("Found occupancies")

            curr_run["occupancies"].append(_process_occupancies(block))

        # SCF Basis set
        elif match := re.search(r" with +(\d) +cut-off energies.", line):

            if Filters.SCF not in to_parse:
                continue

            ncut = int(match.group(1))
            line = ""
            next(castep_file)
            block = Block.from_re(line, castep_file,
                                  "",
                                  "^-+ <-- SCF", n_end=ncut * 3)
            data = get_only(parse_castep_file(block, Filters.HIGH | Filters.SCF))

            scf = data.pop("scf")
            curr_run["bsc_energies"] = data.pop("energies")

            curr_run["scf"], curr_run["bsc_scf"] = scf[-1], scf[:-1]

            curr_run.update(data)

        # Energies
        elif line.startswith("Final energy"):

            logger("Found energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["final_energy"].append(to_type(get_numbers(line)[-1], float))

        elif "Total energy corrected for finite basis set" in line:

            logger("Found energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["final_basis_set_corrected"].append(
                to_type(get_numbers(line)[-1], float))

        elif "0K energy (E-0.5TS)" in line:

            logger("Found estimated 0K energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["est_0K"].append(to_type(get_numbers(line)[-1], float))

        elif line.startswith("(SEDC) Total Energy"):

            logger("Found SEDC energy correction")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["sedc_correction"].append(to_type(get_numbers(line)[-1], float))

        elif line.startswith("Dispersion corrected final energy"):

            logger("Found SEDC final energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["disperson_corrected"].append(
                to_type(get_numbers(line)[-1], float))

        # Free energies
        elif re.match(rf"Final free energy \(E-TS\) += +({REs.EXPFNUMBER_RE})", line):

            logger("Found free energy (E-TS)")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["free_energy"].append(to_type(get_numbers(line)[-1], float))

        # Solvation energy
        elif line.startswith(" Free energy of solvation"):

            if Filters.SOLVATION not in to_parse:
                continue

            logger("Found solvation energy")

            if "energies" not in curr_run:
                curr_run["energies"] = defaultdict(list)

            curr_run["energies"]["solvation"].append(*to_type(get_numbers(line), float))

        # Spin densities
        elif match := REs.INTEGRATED_SPIN_DENSITY_RE.match(line):

            if Filters.SCF not in to_parse and Filters.SPIN not in to_parse:
                continue

            logger("Found spin")

            val = match["val"] if len(match["val"].split()) == 1 else match["val"].split()

            if "|" in line:
                curr_run["modspin"].append(to_type(val, float))
            else:
                curr_run["spin"].append(to_type(val, float))

        # Initial cell
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Unit Cell"), REs.EMPTY, n_end=3):

            if Filters.CELL not in to_parse:
                continue

            logger("Found cell")

            curr_run["initial_cell"] = _process_unit_cell(block)

        # Cell Symmetry and contstraints
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Symmetry and Constraints"),
                                    "Cell constraints are"):

            if Filters.SYMMETRIES not in to_parse:
                continue

            logger("Found symmetries")

            curr_run["symmetries"], curr_run["constraints"] = _process_symmetry(block)

        # TSS (must be ahead of initial pos)
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("(Reactant|Product)", "x"),
                                    gen_table_re("", "x+"), n_end=2):

            if Filters.TSS not in to_parse or Filters.POSITION not in to_parse:
                continue

            mode = "reactant" if "Reactant" in line else "product"

            logger("Found %s initial states", mode)

            curr_run[mode] = _process_atreg_block(block)

        # Initial pos
        elif block := Block.from_re(line, castep_file,  # Labelled
                                    r"Fractional coordinates of atoms\s+User-defined",
                                    gen_table_re("", "x+")):

            if Filters.POSITION not in to_parse:
                continue

            if "labels" not in curr_run:
                curr_run["labels"] = defaultdict(dict)

            logger("Found initial positions")

            curr_run["initial_positions"] = {}
            for line in block:
                if match := REs.LABELLED_POS_RE.search(line):
                    ind = atreg_to_index(match)

                    if lab := match["label"].strip():
                        curr_run["labels"][ind] = lab
                        ind = (f"{ind[0]} [{lab}]", ind[1])
                    else:
                        curr_run["labels"][ind] = "NULL"

                    curr_run["initial_positions"][ind] = to_type(match.group("x", "y", "z"), float)

        elif block := Block.from_re(line, castep_file,  # Mixture
                                    r"Mixture\s+Fractional coordinates of atoms",
                                    gen_table_re("", "x+")):

            if Filters.POSITION not in to_parse:
                continue

            logger("Found initial positions")

            curr_run["initial_positions"] = {}
            for line in block:
                if match := REs.MIXTURE_LINE_1_RE.search(line):
                    val = match.groupdict()
                    spec, idx = atreg_to_index(match)
                    pos = to_type(match.group("x", "y", "z"), float)
                    weight = float(match["weight"])

                    curr_run["initial_positions"][spec, idx] = {"pos": pos, "weight": weight}

                elif match := REs.MIXTURE_LINE_2_RE.search(line):
                    spec = match["spec"].strip()
                    weight = float(match["weight"])

                    curr_run["initial_positions"][spec, idx] = {"pos": pos, "weight": weight}

        elif block := Block.from_re(line, castep_file,
                                    "Fractional coordinates of atoms",
                                    gen_table_re("", "x+")):

            if Filters.POSITION not in to_parse:
                continue

            logger("Found initial positions")

            curr_run["initial_positions"] = _process_atreg_block(block)

        elif "Supercell generated" in line:
            accum = iter(get_numbers(line))
            curr_run["supercell"] = tuple(to_type([next(accum) for _ in range(3)], float)
                                          for _ in range(3))

        # Initial vel
        elif block := Block.from_re(line, castep_file,
                                    "User Supplied Ionic Velocities",
                                    gen_table_re("", "x+")):

            if Filters.POSITION not in to_parse:
                continue

            logger("Found initial velocities")

            curr_run["initial_velocities"] = _process_atreg_block(block)

        # Initial spins
        elif block := Block.from_re(line, castep_file,
                                    "Initial magnetic",
                                    gen_table_re("", "x+")):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found initial spins")

            curr_run["initial_spins"] = _process_initial_spins(block)

        # Target Stress
        elif block := Block.from_re(line, castep_file, "External pressure/stress", "", n_end=3):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found target stress")

            accum = [to_type(number, float)
                     for line in block
                     for number in get_numbers(line)]

            curr_run["target_stress"].append(accum)

        # Finite basis correction parameter
        elif match := re.search(rf"finite basis dEtot\/dlog\(Ecut\) = +({REs.FNUMBER_RE})", line):

            logger("Found dE/dlog(E)")
            curr_run["dedlne"] = to_type(match.group(1), float)

        # K-Points
        elif block := Block.from_re(line, castep_file, "k-Points For BZ Sampling", REs.EMPTY):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found k-points")

            curr_run["k-points"] = _process_kpoint_blocks(block, implicit_kpoints=True)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Number +Fractional coordinates +Weight", r"\+"),
                                    gen_table_re("", r"\++")):

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found k-points list")

            curr_run["k-points"] = _process_kpoint_blocks(block, implicit_kpoints=False)

        elif "Applied Electric Field" in line:

            if Filters.PARAMETERS not in to_parse:
                continue

            logger("Found electric field")
            line = next(castep_file)
            curr_run["applied_field"] = to_type(get_numbers(line), float)

        # Forces blocks
        elif block := Block.from_re(line, castep_file, REs.FORCES_BLOCK_RE, r"^\s*\*+$"):

            if Filters.FORCE not in to_parse:
                continue

            if "forces" not in curr_run:
                curr_run["forces"] = defaultdict(list)

            key, val = _process_forces(block)

            logger("Found %s forces", key)

            curr_run["forces"][key].append(val)

        # Stress tensor block
        elif block := Block.from_re(line, castep_file, REs.STRESSES_BLOCK_RE, r"^\s*\*+$"):
            if Filters.STRESS not in to_parse:
                continue

            if "stresses" not in curr_run:
                curr_run["stresses"] = defaultdict(list)

            key, val = _process_stresses(block)

            logger("Found %s stress", key)

            curr_run["stresses"][key].append(val)

            # Phonon block
        elif block := Block.from_re(line, castep_file,
                                    "Vibrational Frequencies",
                                    gen_table_re("", "=+")):

            if Filters.PHONON not in to_parse:
                continue

            logger("Found phonon")

            curr_run["phonons"] = _process_phonon(block, logger)

            logger("Found %d phonon samples", len(curr_run["phonons"]))

        # Phonon Symmetry
        elif block := Block.from_re(line, castep_file,
                                    "Phonon Symmetry Analysis", REs.EMPTY):

            if Filters.PHONON not in to_parse:
                continue

            logger("Found phonon symmetry analysis")

            val = _process_phonon_sym_analysis(block)
            curr_run["phonon_symmetry_analysis"].append(val)

        # Dynamical Matrix
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Dynamical matrix"),
                                    gen_table_re("", "-+")):

            if Filters.PHONON not in to_parse:
                continue

            logger("Found dynamical matrix")

            val = _process_dynamical_matrix(block)
            curr_run["dynamical_matrix"] = val

        # Raman tensors
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Raman Susceptibility Tensors[^+]*", r"\+"),
                                    REs.EMPTY):

            if Filters.PHONON not in to_parse:
                continue

            logger("Found Raman")

            curr_run["raman"].append(_process_raman(block))

        # Solvation
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("AUTOSOLVATION CALCULATION RESULTS", r"\*+"),
                                    r"^\s*\*+\s*$"):

            if Filters.SOLVATION not in to_parse:
                continue

            logger("Found autosolvation")

            curr_run["autosolvation"] = _process_autosolvation(block)

        # Permittivity and NLO Susceptibility
        elif block := Block.from_re(line, castep_file,
                                    r"^\s+Optical Permittivity", r"^ =+$"):

            if Filters.OPTICS not in to_parse:
                continue

            logger("Found optical permittivity")

            val = _process_3_6_matrix(block, split=True)
            curr_run["optical_permittivity"] = val[0]
            if val[1]:
                curr_run["dc_permittivity"] = val[1]

        # Polarisability
        elif block := Block.from_re(line, castep_file, r"^\s+Polarisabilit(y|ies)", r"^ =+$"):

            if Filters.OPTICS not in to_parse:
                continue

            logger("Found polarisability")

            val = _process_3_6_matrix(block, split=True)
            curr_run["optical_polarisability"] = val[0]
            if val[1]:
                curr_run["static_polarisability"] = val[1]

        # Non-linear
        elif block := Block.from_re(line, castep_file,
                                    r"^\s+Nonlinear Optical Susceptibility", r"^ =+$"):

            if Filters.OPTICS not in to_parse:
                continue

            logger("Found NLO")

            curr_run["nlo"], _ = _process_3_6_matrix(block, split=False)

        # Atomic displacements
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re(r"Atomic Displacement Parameters \(A\*\*2\)"),
                                    gen_table_re("", "-+"), n_end=3):

            if Filters.THERMODYNAMICS not in to_parse:
                continue

            logger("Found atomic displacements")

            accum = _process_atom_disp(block)
            curr_run["atomic_displacements"] = accum

        # Thermodynamics
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Thermodynamics"),
                                    gen_table_re("", "-+"), n_end=3):

            if Filters.THERMODYNAMICS not in to_parse:
                continue

            logger("Found thermodynamics")

            accum = _process_thermodynamics(block)
            curr_run["thermodynamics"] = accum

        # Mulliken Population Analysis
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re(r"Atomic Populations \(Mulliken\)"),
                                    gen_table_re("", "=+"), n_end=2):

            if Filters.POPN_ANALYSIS not in to_parse:
                continue

            logger("Found Mulliken")

            curr_run["mulliken_popn"] = _process_mulliken(block)

        # Born charges
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Born Effective Charges"),
                                    gen_table_re("", "=+")):

            if Filters.POPN_ANALYSIS not in to_parse:
                continue

            logger("Found Born")

            curr_run["born"].append(_process_born(block))

        # Orbital populations
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Orbital Populations"),
                                    gen_table_re("", "-+"), n_end=3):

            if Filters.POPN_ANALYSIS not in to_parse:
                continue

            logger("Found Orbital populations")

            curr_run["orbital_popn"] = _process_orbital_populations(block)

        # Bond analysis
        elif block := Block.from_re(line, castep_file,
                                    r"Bond\s+Population(?:\s+Spin)?\s+Length",
                                    gen_table_re("", "=+"), n_end=2):

            if Filters.POPN_ANALYSIS not in to_parse:
                continue

            logger("Found bond info")

            curr_run["bonds"] = _process_bond_analysis(block)

        # Hirshfeld Population Analysis
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Hirshfeld Analysis"),
                                    gen_table_re("", "=+"), n_end=2):

            if Filters.POPN_ANALYSIS not in to_parse:
                continue

            logger("Found Hirshfeld")

            curr_run["hirshfeld"] = _process_hirshfeld(block)

        # ELF
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("ELF grid sample"),
                                    gen_table_re("", "-+"), n_end=2):

            if Filters.ELF not in to_parse:
                continue

            logger("Found ELF")

            curr_run["elf"] = _process_elf(block)

        # MD Block
        elif ((block := Block.from_re(line, castep_file,  # Capture general MD step
                                    "Starting MD iteration",
                                     "finished MD iteration")) or
              (block := Block.from_re(line, castep_file,  # Capture 0th iteration
                                     "Starting MD",
                                     gen_table_re("", "=+")))):

            if Filters.MD not in to_parse:
                continue

            logger("Found MD Block (step %d)", len(curr_run["md"]))

            # Avoid infinite recursion
            next(block)
            data = get_only(parse_castep_file(block))
            add_aliases(data, {"initial_positions": "positions",
                               "initial_cell": "cell"},
                        replace=True)

            # Put memory estimate to top level
            if "memory_estimate" in data:
                curr_run["memory_estimate"] = data.pop("memory_estimate")

            curr_run["md"].append(data)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("MD Data:", "x"),
                                    gen_table_re("", "x+")):

            if Filters.MD not in to_parse and Filters.MD_SUMMARY not in to_parse:
                continue

            curr_run.update(_process_md_block(block))

        # GeomOpt
        elif block := Block.from_re(line, castep_file, "Final Configuration",
                                    rf"\s*{REs.MINIMISERS_RE}: Final"):

            if Filters.GEOM_OPT not in to_parse and Filters.FINAL_CONFIG not in to_parse:
                continue

            logger("Found final geom configuration")

            curr_run["geom_opt"]["final_configuration"] = _process_final_config_block(block)

        elif block := Block.from_re(line, castep_file,
                                    rf"Starting {REs.MINIMISERS_RE} iteration\s*\d+\s*\.{{3}}",
                                    rf"^=+$|^\s*Finished\s+{REs.MINIMISERS_RE}\s*$", n_end=2):

            if Filters.GEOM_OPT not in to_parse:
                continue

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            if not curr_run["geom_opt"]["iterations"]:
                data = {key: val for key, val in curr_run.items()
                        if key in {"enthalpy", "initial_cell", "initial_positions",
                                   "scf", "forces", "stresses", "minimisation"}}

                for key in ("enthalpy", "scf", "forces", "stresses", "minimisation"):
                    if key in curr_run:
                        del curr_run[key]

                add_aliases(data, {"initial_positions": "positions",
                                   "initial_cell": "cell"},
                            replace=True)

                curr_run["geom_opt"]["iterations"] = [data]

            logger("Found geom block (iteration %d)", len(curr_run["geom_opt"]["iterations"]) + 1)
            # Avoid infinite recursion
            next(block)
            data = get_only(parse_castep_file(block))

            add_aliases(data, {"initial_positions": "positions",
                               "initial_cell": "cell"},
                        replace=True)
            curr_run["geom_opt"]["iterations"].append(data)

        elif match := re.search(f"(?P<minim>{REs.MINIMISERS_RE}):"
                                r" finished iteration\s*\d+\s*with enthalpy", line):

            if Filters.GEOM_OPT not in to_parse:
                continue

            minim = match["minim"]

            logger("Found %s energy", minim)

            curr_run["enthalpy"].append(to_type(get_numbers(line)[-1], float))

        elif match := re.search(rf"trial guess \(lambda=\s*({REs.EXPFNUMBER_RE})\)", line):

            if "geom_opt" not in curr_run:
                curr_run["geom_opt"] = defaultdict(list)

            curr_run["geom_opt"]["trial"].append(float(match.group(1)))

        elif match := re.match(rf"^\s*(?:{REs.MINIMISERS_RE}):\s*"
                               r"(?P<key>Final [^=]+)=\s*"
                               f"(?P<value>{REs.EXPFNUMBER_RE}).*",
                               line, re.IGNORECASE):

            if Filters.GEOM_OPT not in to_parse:
                continue

            key, val = normalise_string(match["key"]).lower(), to_type(match["value"], float)
            key = "_".join(key.split())
            logger("Found geomopt %s", key)
            curr_run["geom_opt"]["final_configuration"][key] = val

        elif block := Block.from_re(line, castep_file,
                                    f"<--( min)? {REs.MINIMISERS_RE}$",
                                    r"\+(?:-+\+){4,5}", n_end=2):

            if Filters.GEOM_OPT not in to_parse:
                continue

            if not (match := re.search(REs.MINIMISERS_RE, line)):
                raise ValueError("Invalid Geom block")

            typ = match.group(0)

            logger("Found %s geom_block", typ)

            curr_run["minimisation"].append(_process_geom_table(block))

        # GeomOpt Deloc

        elif block := Block.from_re(line, castep_file,
                                    "INTERNAL CONSTRAINTS", REs.EMPTY, n_end=2):

            if Filters.GEOM_OPT not in to_parse:
                continue

            curr_run["internal_constraints"] = _process_internal_constraints(block)

        elif block := Block.from_re(line, castep_file,
                                    "Message: Generating deloc", "Message: Generation of deloc"):

            if Filters.GEOM_OPT not in to_parse:
                continue

            logger("Found Delocalisaed GeomOpt Table")

            curr_run.setdefault("delocalised_internal", {})
            curr_run["delocalised_internal"].update(_process_deloc_table(block))

        elif block := Block.from_re(line, castep_file,
                                    "The size of active space", REs.EMPTY):

            if Filters.GEOM_OPT not in to_parse:
                continue

            logger("Found Delocalisaed GeomOpt")

            curr_run.setdefault("delocalised_internal", {})
            curr_run["delocalised_internal"].update(_process_deloc_act_space_table(block))

        # TDDFT
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("TDDFT excitation energies", r"\+", post="TDDFT"),
                                    gen_table_re("=+", r"\+", post="TDDFT"), n_end=2):

            if Filters.TDDFT not in to_parse:
                continue

            logger("Found TDDFT excitations")

            curr_run["tddft"] = _process_tddft(block)

        # Band structure
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("(B A N D|Band Structure Calculation)[^+]+",
                                                 r"\+"),
                                    gen_table_re("", "=+")):

            if Filters.BS not in to_parse:
                continue

            logger("Found band-structure")

            curr_run["bs"] = _process_band_structure(block)

        # Molecular Dipole
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("D I P O L E   O F   M O L E C U L E"
                                                 "   I N   S U P E R C E L L",
                                                 r"\+"),
                                    gen_table_re("", "=+")):

            if Filters.DIPOLE not in to_parse:
                continue

            logger("Found molecular dipole")

            curr_run["molecular_dipole"] = _process_dipole(block)

        # Chemical shielding
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Chemical Shielding Tensor", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.CHEM_SHIELDING not in to_parse:
                continue

            logger("Found Chemical Shielding Tensor")

            val = _parse_magres_block(0, block)
            curr_run["magres"].append(val)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Chemical Shielding and "
                                                 "Electric Field Gradient Tensors", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.CHEM_SHIELDING not in to_parse:
                continue

            logger("Found Chemical Shielding + EField Tensor")

            val = _parse_magres_block(1, block)
            curr_run["magres"].append(val)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Electric Field Gradient Tensor", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.CHEM_SHIELDING not in to_parse:
                continue

            logger("Found EField Tensor")

            val = _parse_magres_block(2, block)
            curr_run["magres"].append(val)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("(?:I|Ani)sotropic J-coupling", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.CHEM_SHIELDING not in to_parse:
                continue

            logger("Found J-coupling")

            val = _parse_magres_block(3, block)
            curr_run["magres"].append(val)

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Hyperfine Tensor", r"\|"),
                                    gen_table_re("", "=+")):

            if Filters.CHEM_SHIELDING not in to_parse:
                continue

            logger("Found Hyperfine tensor")

            val = _parse_magres_block(4, block)
            curr_run["magres"].append(val)

        # Elastic
        elif block := Block.from_re(line, castep_file,
                                    gen_table_re(r"Elastic Constants Tensor \(GPa\)"),
                                    gen_table_re("", "=+")):

            if Filters.ELASTIC not in to_parse:
                continue

            logger("Found elastic constants tensor")

            curr_run.setdefault("elastic", {})

            val, _ = _process_3_6_matrix(block, split=False)
            curr_run["elastic"]["elastic_constants"] = val

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re(r"Compliance Matrix \(GPa\^-1\)"),
                                    gen_table_re("", "=+")):

            if Filters.ELASTIC not in to_parse:
                continue

            logger("Found compliance matrix")

            curr_run.setdefault("elastic", {})

            val, _ = _process_3_6_matrix(block, split=False)
            curr_run["elastic"]["compliance_matrix"] = val

        elif block := Block.from_re(line, castep_file, "Contribution ::", REs.EMPTY):

            if not (match := re.match(r"(?P<type>.* Contribution)", line)):
                raise ValueError("Invalid elastic block")

            typ = match.group("type")
            next(block)

            if Filters.ELASTIC not in to_parse:
                continue

            logger("Found elastic %s contribution", typ)

            curr_run.setdefault("elastic", {})

            val, _ = _process_3_6_matrix(block, split=False)
            curr_run["elastic"][typ] = val

        elif block := Block.from_re(line, castep_file,
                                    gen_table_re("Elastic Properties"),
                                    gen_table_re("", "=+")):

            if Filters.ELASTIC not in to_parse:
                continue

            logger("Found elastic properties")

            curr_run.setdefault("elastic", {})

            curr_run["elastic"].update(_process_elastic_properties(block))

        # Berry phase polarisation

        elif block := Block.from_re(line, castep_file,  # Polarisation verbose
                                    r"^\s*Ionic contribution to polarisation", "=+", n_end=14):

            if Filters.POLARISATION not in to_parse:
                continue

            logger("Found verbose berry phase polarisation")

            curr_run.setdefault("berry_phase", {})

            curr_run["berry_phase"].update(_process_berry_phase(block))

        elif block := Block.from_re(line, castep_file,  # Polarisation
                                    r"^\s*Polarisation", "=+", n_end=6):

            if Filters.POLARISATION not in to_parse:
                continue

            logger("Found berry phase polarisation")

            curr_run.setdefault("berry_phase", {})

            curr_run["berry_phase"].update(_process_berry_phase(block))

        # --- Extra blocks for testing

        elif block := Block.from_re(
            line,
            castep_file,
            f"<BEGIN ({'|'.join(PARSERS)})>",
            f"<END ({'|'.join(PARSERS)})>",
        ):

            if Filters.TEST_EXTRA_DATA not in to_parse:
                continue

            key = re.search(r"BEGIN (\w+)", line).group(1)

            logger("Found %s block", key)

            curr_run.setdefault("external_files", {})

            block.remove_bounds(1, 1)

            val = PARSERS[key](block)
            curr_run["external_files"][key] = val

    if curr_run:
        fix_data_types(curr_run, {"energies": float,
                                  "solvation": float})
        runs.append(curr_run)
    return runs


def _process_ps_energy(block: Block) -> tuple[str, PSPotEnergy]:
    if not (match := REs.PS_SHELL_RE.search(next(block))):
        raise ValueError("Invalid PS Energy")

    key = match["spec"]
    accum: PSPotEnergy = defaultdict(list)

    next(block)
    accum["pseudo_atomic_energy"] = float(get_numbers(next(block))[1])
    for line in block:
        if "Charge spilling" in line:
            assert isinstance(accum["charge_spilling"], list)
            accum["charge_spilling"].append(0.01 * float(get_numbers(line)[-1]))

    if "charge_spilling" in accum:
        accum["charge_spilling"] = tuple(accum["charge_spilling"])

    return key, accum


def _process_tddft(block: Block) -> list[TDDFTData]:
    return [
        {"energy": float(match["energy"]),
         "error": float(match["error"]),
         "type": match["type"]}
        for line in block
        if (match := REs.TDDFT_RE.match(line))
    ]


def _process_atreg_block(block: Block) -> AtomPropBlock:
    return {
        atreg_to_index(match): to_type(match.group("x", "y", "z"), float)
        for line in block
        if (match := REs.ATOMIC_DATA_3VEC.search(line))
    }


def _process_spec_prop(block: Block) -> list[list[str]]:

    accum = []

    for line in block:
        words = line.split()
        if words and re.match(rf"{REs.SPECIES_RE}\b", words[0]):

            accum.append(words)

    return accum


def _process_md_block(block: Block) -> MDInfo:
    curr_data = {match.group("key").strip(): float(match.group("val"))
                 for line in block
                 if (match := re.search(r"x\s+"
                                        r"(?P<key>[a-zA-Z][A-Za-z ]+):\s*"
                                        rf"(?P<val>{REs.FNUMBER_RE})", line))}

    return {normalise_key(key): val for key, val in curr_data.items()}


def _process_elf(block: Block) -> list[float]:
    return [
        float(match.group(1)) for line in block
        if (match := re.match(rf"\s+ELF\s+\d+\s+({REs.FNUMBER_RE})", line))
    ]


def _process_hirshfeld(block: Block) -> dict[AtomIndex, float]:
    """Process Hirshfeld block to dict of charges.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[AtomIndex, float]
        Hirshfeld charges.
    """
    return {
        atreg_to_index(match): float(match["charge"]) for line in block
        if (match := re.match(rf"\s+{REs.ATOM_RE}\s+(?P<charge>{REs.FNUMBER_RE})", line))
    }


def _process_thermodynamics(block: Block) -> Thermodynamics:
    """Process a thermodynamics block into a dict of lists.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    Thermodynamics
        Thermodynamics data.
    """
    accum: Thermodynamics = defaultdict(list)
    for line in block:
        if "Zero-point energy" in line:
            accum["zero_point_energy"] = float(get_numbers(line)[0])

        if match := REs.THERMODYNAMICS_DATA_RE.match(line):
            stack_dict(accum, match.groupdict())

        # elif re.match(r"\s+T\(", line):  # Can make dict/re based on labels
        #     thermo_label = line.split()

    fix_data_types(accum, dict.fromkeys(("t", "e", "f", "s", "cv"), float))
    return accum


def _process_atom_disp(block: Block) -> dict[str, dict[AtomIndex, SixVector]]:
    """Process a atom disp block into a dict of lists.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[str, dict[AtomIndex, SixVector]]
        Atom displacements information.
    """
    accum: dict[str, dict[AtomIndex, SixVector]] = defaultdict(dict)
    for line in block:
        if match := REs.ATOMIC_DISP_RE.match(line):
            ind = atreg_to_index(match.groupdict())
            temp, disp = match["temperature"], match["displacement"]
            val = to_type(disp.split(), float)
            accum[temp][ind] = val

    return accum


def _process_3_6_matrix(
        block: Block, *, split: bool,
) -> tuple[ThreeByThreeMatrix, ThreeByThreeMatrix | None]:
    """Process a single or pair of 3x3 matrices or 3x6 matrix.

    Parameters
    ----------
    block : Block
        Block to parse.
    split : bool
        Whether to try to split 3x6 values into two 3x3 matrices.

    Returns
    -------
    ThreeByThreeMatrix
        First matrix.
    ThreeByThreeMatrix | None
        Second matrix if present or None

    Examples
    --------
    >>> threebythree = Block.from_iterable(('1 2 3', '4 5 6', '7 8 9'))
    >>> a, b = _process_3_6_matrix(threebythree, split=False)
    >>> b is None
    True
    >>> for i in a:
    ...     print(i)
    (1.0, 2.0, 3.0)
    (4.0, 5.0, 6.0)
    (7.0, 8.0, 9.0)
    >>> threebysix = Block.from_iterable(('1 2 3 4 5 6', '7 8 9 10 11 12', '13 14 15 16 17 18'))
    >>> a, b = _process_3_6_matrix(threebysix, split=False)
    >>> b is None
    True
    >>> for i in a:
    ...     print(i)
    (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)
    (7.0, 8.0, 9.0, 10.0, 11.0, 12.0)
    (13.0, 14.0, 15.0, 16.0, 17.0, 18.0)
    >>> threebysix = Block.from_iterable(('1 2 3 4 5 6', '7 8 9 10 11 12', '13 14 15 16 17 18'))
    >>> a, b = _process_3_6_matrix(threebysix, split=True)
    >>> for i in a:
    ...     print(i)
    (1.0, 2.0, 3.0)
    (7.0, 8.0, 9.0)
    (13.0, 14.0, 15.0)
    >>> for i in b:
    ...     print(i)
    (4.0, 5.0, 6.0)
    (10.0, 11.0, 12.0)
    (16.0, 17.0, 18.0)
    """
    parsed = tuple(to_type(vals, float) for line in block
                   if (vals := get_numbers(line)) and len(vals) in {3, 6})

    if split and len(parsed[0]) == 6:
        fst = cast(ThreeByThreeMatrix, tuple(line[0:3] for line in parsed))
        snd = cast(ThreeByThreeMatrix, tuple(line[3:6] for line in parsed))
    else:
        fst = cast(ThreeByThreeMatrix, parsed)
        snd = None

    return fst, snd


def _process_params(block: Block) -> dict[str, dict[str, str | tuple[Any, ...]]]:
    """Process a parameters block into a dict of params.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[str, dict[str, str | tuple[Any, ...]]]
        Parsed info.
    """
    opt: dict[str, Any] = {}
    curr_opt: dict[str, str | tuple[Any, ...]] = {}
    curr_group = ""

    match: re.Match | list[str] | None
    key: str | list[str]
    val: Any

    for line in block:
        if dev_block := Block.from_re(line, block, "Developer Code", gen_table_re("", r"\*+")):

            opt["devel_code"] = _parse_devel_code_block(dev_block)

        elif match := re.match(r"\s*\*+ ([A-Za-z -]+) \*+", line):
            if curr_opt:
                opt[curr_group] = curr_opt

            curr_group = normalise_string(match.group(1).replace("Parameters", "")).lower()
            curr_opt = {}

        elif match := re.match(r"\s*output (?P<key>.*) unit\s*:\s*(?P<val>.*)", line):

            if "output_units" not in opt:
                opt["output_units"] = {}

            key, val = map(normalise_string, match.group("key", "val"))
            opt["output_units"][key] = val

        elif len(match := line.split(":")) > 1:
            *key, val = map(normalise_string, match)
            if val.split() and REs.NUMBER_RE.match(val.split()[0]):
                val, *unit = val.split()
                val = to_type(val, determine_type(val))
                if unit:
                    val = (val, *unit)

            curr_opt[" ".join(key).strip()] = val

    if curr_opt:
        opt[curr_group] = curr_opt

    return opt


def _process_buildinfo(block: Block) -> dict[str, str]:
    info = {}

    info["summary"] = " ".join(map(normalise_string, block[0:2]))
    for line in block[2:]:
        if ":" in line:
            key, val = map(normalise_string, line.split(":", 1))
            info[normalise_key(key)] = val.strip()
    return info


def _process_unit_cell(block: Block) -> CellInfo:
    cell: CellInfo = defaultdict(list)
    prop = []

    assert isinstance(cell["real_lattice"], list)
    assert isinstance(cell["recip_lattice"], list)
    assert isinstance(cell["lattice_parameters"], list)
    assert isinstance(cell["cell_angles"], list)

    for line in block:
        numbers = get_numbers(line)
        if len(numbers) == 6:
            cell["real_lattice"].append(to_type(numbers[0:3], float))
            cell["recip_lattice"].append(to_type(numbers[3:6], float))
        elif len(numbers) == 2:
            if any(ang in line for ang in ("alpha", "beta", "gamma")):
                cell["lattice_parameters"].append(to_type(numbers[0], float))
                cell["cell_angles"].append(to_type(numbers[1], float))
            else:
                prop.append(float(numbers[0]))

    cell.update({name: val for val, name in zip(prop, ("volume", "density_amu", "density_g"))})

    return cell


def _process_scf(block: Block) -> list[SCFReport]:
    scf = []
    curr: SCFReport = {}
    debug_info: SCFDebugInfo = {}
    for line in block:
        if match := REs.SCF_LOOP_RE.match(line):
            if curr:
                scf.append(curr)

            curr = match.groupdict()
            if debug_info:  # Debug info refers to next cycle
                curr["debug_info"] = debug_info
            debug_info = {}
            fix_data_types(curr, {"energy": float,
                                  "energy_gain": float,
                                  "fermi_energy": float,
                                  "time": float})

        elif "Density was not mixed" in line:
            curr["density_residual"] = None

        elif "Norm of density" in line:
            curr["density_residual"] = to_type(get_numbers(line)[0], float)

        elif "constraint energy" in line:
            curr["constraint_energy"] = to_type(get_numbers(line)[0], float)

        elif "Correcting PBC dipole-dipole" in line:
            curr["dipole_corr_energy"] = to_type(get_numbers(line)[0], float)

        # Debug info

        elif "no. bands" in line:
            debug_info["no_bands"] = to_type(get_numbers(line)[0], int)

        elif "Kinetic eigenvalue" in line:
            debug_info.setdefault("kinetic_eigenvalue", [])
            debug_info.setdefault("eigenvalue", [])

            debug_info["kinetic_eigenvalue"].append(to_type(get_numbers(line)[1], float))
            eig = []

        elif re.match(r"eigenvalue\s*\d+\s*init=", line):
            numbers = get_numbers(line)
            eig.append({"initial": float(numbers[1]),
                        "final": float(numbers[2]),
                        "change": float(numbers[3])})

        elif "Checking convergence criteria" in line:
            debug_info["eigenvalue"].append(eig)
            eig = []

        elif match := re.match(r"[+(]?(?P<key>[()0-9A-Za-z -]+)=?"
                               rf"\s*{labelled_floats(('val',))} eV\)?", line):
            debug_info.setdefault("contributions", {})
            key, val = normalise_key(match["key"]), float(match["val"])

            fix_keys = {"apolar_corr_to_eigenvalue_sum": "apolar_correction",
                        "hubbard_u_correction_to_eigenvalu": "hubbard_u_correction",
                        "xc_correction_to_eigenvalue_sum": "xc_correction"}
            key = fix_keys.get(key, key)

            debug_info["contributions"][key] = val

    if curr:
        scf.append(curr)

    return scf


def _process_forces(block: Block) -> tuple[str, AtomPropBlock]:
    if not (ft_guess := REs.FORCES_BLOCK_RE.search(next(block))):
        raise ValueError("Invalid forces block")
    ftype = ft_guess.group(1) or "non_descript"
    ftype = normalise_key(ftype)

    accum = {atreg_to_index(match): to_type(match.group("x", "y", "z"), float)
             for line in block
             if (match := REs.ATOMIC_DATA_FORCE.search(line))}

    return ftype, accum


def _process_stresses(block: Block) -> tuple[str, SixVector]:
    if not (ft_guess := REs.STRESSES_BLOCK_RE.search(next(block))):
        raise ValueError("Invalid stresses block")
    ftype = ft_guess.group(1) or "non_descript"
    ftype = normalise_key(ftype)

    accum = []
    for line in block:
        numbers = get_numbers(line)
        if "*  x" in line:
            accum += numbers[0:]
        elif "*  y" in line:
            accum += numbers[1:]
        elif "*  z" in line:
            accum += numbers[2:]

    return ftype, to_type(accum, float)


def _process_initial_spins(block: Block) -> dict[AtomIndex, InitialSpin]:
    """Process a set of initial spins into appropriate dict.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[AtomIndex, InitialSpin]
        Parsed info.
    """
    accum: dict[AtomIndex, InitialSpin] = {}
    for line in block:
        if match := re.match(rf"\s*\|\s*{REs.ATOM_RE}\s*"
                             rf"{labelled_floats(('spin', 'magmom'))}\s*"
                             r"(?P<fix>[TF])\s*\|", line):
            val = match.groupdict()
            ind = atreg_to_index(val)
            fix_data_types(val, {"spin": float, "magmom": float})
            val["fix"] = val["fix"] == "T"
            accum[ind] = cast(dict[str, Union[float, bool]], val)
    return accum


def _process_born(block: Block) -> dict[AtomIndex, ThreeByThreeMatrix]:
    """Process a Born block into a dict of charges.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[AtomIndex, ThreeByThreeMatrix]
        Parsed info.
    """
    born_accum = {}
    for line in block:
        if match := REs.BORN_RE.match(line):
            val = match.groupdict()
            label = val.pop("label")
            if label is not None:
                val["spec"] = f"{val['spec']} [{label}]"

            born_accum[atreg_to_index(val)] = (to_type(val["charges"].split(), float),
                                               to_type(next(block).split(), float),
                                               to_type(next(block).split(), float))
    return born_accum


def _process_raman(block: Block) -> list[RamanReport]:
    """Process a Mulliken block into a list of modes.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    list[RamanReport]
        Parsed info.
    """
    next(block)  # Skip first captured line
    modes = []
    curr_mode: RamanReport = {}
    for line in block:
        if "Mode number" in line:
            if curr_mode:
                modes.append(curr_mode)

            curr_mode = {"tensor": [], "depolarisation": None}
        elif numbers := get_numbers(line):
            assert isinstance(curr_mode["tensor"], list)
            curr_mode["tensor"].append(to_type(numbers[0:3], float))
            if len(numbers) == 4:
                curr_mode["depolarisation"] = float(numbers[3])

        elif re.search(r"^ \+\s+\+", line):  # End of 3x3+depol block
            # Compute Invariants Tr(A) and Tr(A)^2-Tr(A^2) of Raman Tensor
            assert isinstance(curr_mode["tensor"], list)
            tensor = curr_mode["tensor"]
            curr_mode["tensor"] = cast(ThreeByThreeMatrix, tuple(curr_mode["tensor"]))
            curr_mode["trace"] = sum(tensor[i][i] for i in range(3))
            curr_mode["ii"] = (tensor[0][0] * tensor[1][1] +
                               tensor[0][0] * tensor[2][2] +
                               tensor[1][1] * tensor[2][2] -
                               tensor[0][1] * tensor[1][0] -
                               tensor[0][2] * tensor[2][0] -
                               tensor[1][2] * tensor[2][1])
    if curr_mode:
        modes.append(curr_mode)

    return modes


def _process_mulliken(block: Block) -> dict[AtomIndex, MullikenInfo]:
    """Process a mulliken block into a dict of points.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    dict[AtomIndex, MullikenInfo]
        Parsed info.

    Raises
    ------
    ValueError
        On malformed "down" spin.
    """
    accum = {}

    for line in block:
        if match := REs.POPN_RE.match(line):
            mull = match.groupdict()
            mull["spin_sep"] = bool(mull["spin_sep"])
            if mull["spin_sep"]:  # We have spin separation
                add_aliases(mull,
                            {orb: f"up_{orb}" for orb in (*SHELLS, "total")},
                            replace=True)
                line = next(block)
                if not (match := REs.POPN_RE_DN.match(line)):
                    raise ValueError("Invalid mulliken down spin")
                val = match.groupdict()

                add_aliases(val,
                            {orb: f"dn_{orb}" for orb in (*SHELLS, "total")},
                            replace=True)

                mull.update(val)
                mull["total"] = float(mull["up_total"]) + float(mull["dn_total"])

            ind = atreg_to_index(mull)
            fix_data_types(mull, {**{f"{orb}": float for orb in (*SHELLS, "total",
                                                                 "charge", "spin")},
                                  **{f"up_{orb}": float for orb in (*SHELLS, "total")},
                                  **{f"dn_{orb}": float for orb in (*SHELLS, "total")}})
            accum[ind] = mull

    return accum


def _process_band_structure(block: Block) -> list[BandStructure]:
    """Process a band structure into a list of kpts.

    Parameters
    ----------
    block : Block
        Block to parse.

    Returns
    -------
    list[BandStructure]
        Parsed info.
    """
    def fdt(qdat: dict) -> None:
        fix_data_types(qdat, {"spin": int,
                              "kx": float,
                              "ky": float,
                              "kz": float,
                              "kpgrp": int,
                              "band": int,
                              "energy": float})

    bands = []
    qdata: BandStructure = {}

    for line in block:
        if match := REs.BS_RE.search(line):
            if qdata:
                fdt(qdata)
                bands.append(qdata)

            qdata = {"band": [], "energy": [], **match.groupdict()}

        elif match := re.search(labelled_floats(("band", "energy"), sep=r"\s+"), line):
            stack_dict(qdata, match.groupdict())

    if qdata:
        fdt(qdata)
        bands.append(qdata)

    return bands


def _process_qdata(qdata: dict[str, str | list[str]]) -> QData:
    """Parse phonon qdata into components.

    Remove empty values and qpt then fix types.

    Parameters
    ----------
    qdata : dict[str, str | list[str]]
        Dictionary containing q-data.

    Returns
    -------
    QData
        Corrected q-data
    """
    qdata = {key: val
             for key, val in qdata.items()
             if any(val) or key == "qpt"}
    fix_data_types(qdata,
                   {"qpt": float,
                    "N": int,
                    "frequency": float,
                    "ir_intensity": float,
                    "raman_intensity": float,
                    })
    return cast(QData, qdata)


def _parse_magres_block(task: int, inp: Block) -> dict[str | AtomIndex,
                                                       str | dict[str, float | None]]:
    """Parse MagRes data tables from inp according to task.

    Parameters
    ----------
    task : int
        Current magres task.
    inp : Block
        Block to parse.

    Returns
    -------
    dict[str | AtomIndex, str | dict[str, float | None]]
        Parsed info.
    """
    data: dict[str | AtomIndex, str | dict[str, float | None]] = {}
    data["task"] = REs.MAGRES_TASK[task]
    curr_re = REs.MAGRES_RE[task]
    for line in inp:
        if match := curr_re.match(line):
            tmp = match.groupdict()
            ind = atreg_to_index(tmp)

            val: dict[str, float | None] = {key: float(val)
                                               for key, val in tmp.items() if key != "asym"}
            if "asym" in tmp:
                val["asym"] = float(tmp["asym"]) if tmp["asym"] != "N/A" else None

            data[ind] = val

    return data


def _process_finalisation(block: Block) -> dict[str, float]:

    out = {}

    for line in block:
        if line.strip():
            key, val = line.split("=")
            out[normalise_key(key)] = float(get_numbers(val)[0])
    return out


def _process_memory_est(block: Block) -> dict[str, MemoryEst]:

    accum = {}

    for line in block:
        if match := re.match(r"\s*\|([A-Za-z ]+)" +
                             labelled_floats(("memory", "disk"), suffix=" MB"), line):
            key, memory, disk = match.groups()
            accum[normalise_key(key)] = {"memory": float(memory),
                                         "disk": float(disk)}

    return accum


def _process_phonon_sym_analysis(block: Block) -> PhononSymmetryReport:
    accum: PhononSymmetryReport = {"title": "", "mat": ()}
    accum["title"] = normalise_string(next(block).split(":")[1])
    next(block)
    accum["mat"] = tuple(parse_int_or_float(numbers)
                         for line in block
                         if (numbers := get_numbers(line)))
    return accum


def _process_kpoint_blocks(block: Block, *,
                           implicit_kpoints: bool) -> KPointsList | KPointsSpec:

    if implicit_kpoints:
        accum: KPointsSpec = {}
        for line in block:
            if "MP grid size" in line:
                accum["kpoint_mp_grid"] = to_type(get_numbers(line), int)
            elif "offset" in line:
                accum["kpoint_mp_offset"] = to_type(get_numbers(line), float)
            elif "Number of kpoints" in line:
                accum["num_kpoints"] = to_type(get_numbers(line)[0], int)
    else:
        kp_block_re = re.compile(gen_table_re(r"\d\s*" + labelled_floats(("qx", "qy", "qz", "wt")),
                                              r"\+"))
        accum: KPointsList = {"points": [{"qpt": to_type(match.group("qx", "qy", "qx"), float),
                                          "weight": to_type(match["wt"], float)}
                                         for line in block
                                         if (match := kp_block_re.match(line))]}
        assert isinstance(accum["points"], list)
        accum["num_kpoints"] = len(accum["points"])

    return accum


def _process_symmetry(block: Block) -> tuple[SymmetryReport, ConstraintsReport]:

    sym: dict[str, Any] = {}
    con: dict[str, Any] = {}
    val: Any

    for line in block:
        if "=" in line:
            key, val = line.split("=")
            key = normalise_key(key)
            val = normalise_string(val)

            if "Number of" in line:
                val = to_type(val, int)

            if "constraints" in key:
                con[key] = val
            else:
                sym[key] = val

        elif re.match(r"\s*\d+\s*rotation\s*", line):
            if "symop" not in sym:
                sym["symop"] = []

            curr_sym: dict[str, Any] = {"rotation": [], "symmetry_related": []}
            for curr_ln in itertools.islice(block, 3):
                curr_sym["rotation"].append(to_type(get_numbers(curr_ln), float))

            next(block)  # elif re.match(r"\s*\d+\s*displacement\s*", line):

            curr_sym["displacement"] = to_type(get_numbers(next(block)), float)

            next(block)  # elif "symmetry related atoms:" in line:

            while line := next(block).strip():
                key, val = line.split(":")
                curr_sym["symmetry_related"].extend((key, int(ind))
                                                    for ind in val.split())

            sym["symop"].append(curr_sym)

        elif "Cell is a supercell containing" in line:

            sym["n_primitives"] = to_type(get_numbers(line)[0], int)

        elif "Centre of mass" in line:
            con["com_constrained"] = "NOT" not in line

        elif cons_block := Block.from_re(line, block, r"constraints\.{5}", r"\s*x+\.{4}\s*"):
            con["ionic_constraints"] = defaultdict(list)
            for match in re.finditer(rf"{REs.ATOM_RE}\s*[xyz]\s*" +
                                     labelled_floats(("pos",), counts=(3,)),
                                     str(cons_block)):
                val = match.groupdict()
                ind = atreg_to_index(val)
                con["ionic_constraints"][ind].append(to_type(val["pos"].split(), float))

        elif "Cell constraints are:" in line:
            con["cell_constraints"] = to_type(get_numbers(line), int)

    return sym, con


def _process_dynamical_matrix(block: Block) -> tuple[tuple[complex, ...], ...]:
    next(block)  # Skip header line
    next(block)

    real_part = []
    while (line := next(block)) and "Ion" not in line:
        numbers = get_numbers(line)
        real_part.append(numbers[2:])

    # Get remainder
    imag_part = [numbers[2:] for line in block if (numbers := get_numbers(line))]

    return tuple(
        tuple(complex(float(real), float(imag)) for real, imag in zip(real_row, imag_row))
        for real_row, imag_row in zip(real_part, imag_part)
    )


def _process_pspot_report(block: Block) -> PSPotReport:

    accum: PSPotReport = {"reference_electronic_structure": [],
                          "pseudopotential_definition": []}

    for line in block:
        if match := REs.PSPOT_REFERENCE_STRUC_RE.match(line):
            val = match.groupdict()
            fix_data_types(val, {"occupation": float, "energy": float})
            accum["reference_electronic_structure"].append(val)
        elif match := REs.PSPOT_DEF_RE.match(line):
            val = match.groupdict()
            # Account for "loc"
            val["beta"] = int(val["beta"]) if val["beta"].isnumeric() else val["beta"]
            fix_data_types(val, {"l": int, "j": int,
                                 "e": float, "Rc": float, "norm": int})
            accum["pseudopotential_definition"].append(val)
        elif match := re.search(rf"Element: (?P<element>{REs.SPECIES_RE})\s+"
                                rf"Ionic charge: (?P<ionic_charge>{REs.FNUMBER_RE})\s+"
                                r"Level of theory: (?P<level_of_theory>[\w\d]+)", line):
            val = match.groupdict()
            val["ionic_charge"] = float(val["ionic_charge"])
            accum.update(val)

        elif match := re.search(r"Atomic Solver:\s*(?P<solver>[\w\s-]+)", line):
            accum["solver"] = normalise_string(match["solver"])

        elif match := REs.PSPOT_RE.search(line):
            accum["detail"] = _parse_pspot_string(match.group(0))

        elif "Augmentation charge Rinner" in line:
            accum["augmentation_charge_rinner"] = to_type(get_numbers(line), float)

        elif "Partial core correction Rc" in line:
            accum["partial_core_correction"] = to_type(get_numbers(line), float)

    return accum


def _process_bond_analysis(block: Block) -> BondData:
    accum = {((match["spec1"], int(match["ind1"])),
              (match["spec2"], int(match["ind2"]))):
             {"population": float(match["population"]),
              "spin": float(match["spin"]) if match["spin"] else None,
              "length": float(match["length"])}
             for line in block
             if (match := REs.BOND_RE.match(line))}

    for val in accum.values():
        if not val["spin"]:
            del val["spin"]

    return accum


def _process_orbital_populations(block: Block) -> dict[str | AtomIndex, Any]:

    accum: dict[str | AtomIndex, Any] = defaultdict(dict)
    for line in block:
        if match := REs.ORBITAL_POPN_RE.match(line):
            val = match.groupdict()
            ind = atreg_to_index(val)
            accum[ind][val["orb"]] = to_type(val["charge"], float)
        elif match := re.match(rf"\s*Total:\s*{labelled_floats(('charge',))}", line):
            accum["total"] = float(match["charge"])
        elif "total projected" in line:
            accum["total_projected"] = to_type(get_numbers(line), float)

    return accum


def _process_dftd(block: Block) -> dict[str, Any]:
    dftd: dict[str, Any] = {"species": {}}
    match: list[str] | re.Match | None
    val: Any

    for line in block:
        if len(match := line.split(":")) == 2:
            key, val = match
            val = normalise_string(val)
            if "Parameter" in key:
                val = to_type(get_numbers(val)[0], float)

            dftd[normalise_key(key)] = val

        elif match := re.match(rf"\s*x\s*(?P<spec>{REs.ATOM_NAME_RE})\s*" +
                               labelled_floats(("C6", "R0")), line):
            dftd["species"][match["spec"]] = {"c6": float(match["C6"]),
                                              "r0": float(match["R0"])}

    return dftd


def _process_occupancies(block: Block) -> list[Occupancies]:
    label = ("band", "eigenvalue", "occupancy")

    accum = [dict(zip(label, numbers)) for line in block if (numbers := get_numbers(line))]
    for elem in accum:
        fix_data_types(elem, {"band": int,
                              "eigenvalue": float,
                              "occupancy": float})
    return cast(list[Occupancies], accum)


def _process_wvfn_line_min(block: Block) -> WvfnLineMin:
    accum: WvfnLineMin = {}
    for line in block:
        if "initial" in line:
            accum["init_energy"], accum["init_de_dstep"] = to_type(get_numbers(line), float)
        elif line.strip().startswith("| step"):
            accum["steps"] = to_type(get_numbers(line), float)
        elif line.strip().startswith("| gain"):
            accum["gain"] = to_type(get_numbers(line), float)

    return accum


def _process_autosolvation(block: Block) -> dict[str, float]:

    accum = {}
    for line in block:
        if len(match := line.split("=")) > 1:
            key = normalise_key(match[0])
            val = cast(float, to_type(get_numbers(line)[0], float))
            accum[key] = val

    return accum


def _process_phonon(block: Block, logger: Logger) -> list[QData]:
    qdata: dict[str, Any] = defaultdict(list)
    accum: list[QData] = []

    for line in block:
        if match := REs.PHONON_RE.match(line):
            if qdata["qpt"] and qdata["qpt"] not in (phonon["qpt"]
                                                     for phonon in accum):
                accum.append(_process_qdata(qdata))

            qdata = defaultdict(list)
            qdata["qpt"] = match["qpt"].split()

            logger("Reading qpt %s", qdata["qpt"], level="debug")

        elif match := REs.PROCESS_PHONON_RE.match(line):
            # ==By mode
            # qdata["modes"].append(match.groupdict())
            # ==By prop
            stack_dict(qdata, match.groupdict())

        elif char_table := Block.from_re(line, block,
                                         r"Rep\s+Mul", gen_table_re("[-=]+", r"\+"),
                                         eof_possible=True):
            headers = next(char_table).split()[4:]
            next(char_table)
            char: list[CharTable] = []
            for char_line in char_table:
                if re.search(r"[-=]{4,}", char_line):
                    break

                head, tail = char_line.split("|")
                _, rep, *name, mul = head.split()
                *vals, _ = tail.split()
                char.append({"chars": tuple(zip(headers, map(int, vals))),
                             "mul": int(mul),
                             "rep": rep,
                             "name": name})

            for row in char:
                if not row["name"]:
                    del row["name"]
                else:
                    row["name"] = row["name"][0]

            qdata["char_table"] = char

    if qdata["qpt"] and qdata["qpt"] not in (phonon["qpt"] for phonon in accum):
        accum.append(_process_qdata(qdata))

    return accum


def _process_dipole(block: Block) -> DipoleTable:

    accum: DipoleTable = {}

    for line in block:
        if match := re.search(r"Total\s*(?P<type>\w+)", line):
            accum[f"total_{match['type']}"] = float(get_numbers(line)[0])

        elif "Centre" in line:
            key = "centre_electronic" if "elec" in line else "centre_positive"
            accum[key] = cast(ThreeVector, to_type(get_numbers(next(block)), float))

        elif "Magnitude" in line:
            accum["dipole_magnitude"] = cast(float, to_type(get_numbers(line)[0], float))

        elif "Direction" in line:
            accum["dipole_direction"] = cast(ThreeVector, to_type(get_numbers(line), float))

    return accum


def _process_pair_params(block_in: Block) -> dict[str, dict[str, dict | str]]:

    accum: dict[str, Any] = {}
    for line in block_in:
        # Two-body
        if block := Block.from_re(line, block_in, "Two Body", r"^\w*\s*\*+\s*$"):
            for blk_line in block:
                if REs.PAIR_POT_RES["two_body_spec"].search(blk_line):
                    matches = REs.PAIR_POT_RES["two_body_spec"].finditer(blk_line)
                    labels = tuple(match.groups() for match in matches)

                elif match := REs.PAIR_POT_RES["two_body_val"].match(blk_line):
                    tag, typ, lab = match.group("tag", "type", "label")
                    if tag:
                        typ = f"{tag}_{typ}"
                    if typ not in accum:
                        accum[typ] = {}
                    if lab not in accum[typ]:
                        accum[typ][lab] = {}

                    accum[typ][lab].update(zip(labels,
                                               to_type(match["params"].split(),
                                                       float)))

                elif match := REs.PAIR_POT_RES["two_body_one_spec"].match(blk_line):
                    labels = ((match["spec"],),)

        # Three-body
        elif block := Block.from_re(line, block_in, "Three Body", r"^\s*\*+\s*$"):
            for blk_line in block:
                if match := REs.PAIR_POT_RES["three_body_spec"].match(blk_line):
                    labels = (tuple(match["spec"].split()),)

                elif match := REs.PAIR_POT_RES["three_body_val"].match(blk_line):
                    tag, typ, lab = match.group("tag", "type", "label")

                    if tag:
                        typ = f"{tag}_{typ}"
                    if typ not in accum:
                        accum[typ] = {}
                    if lab not in accum[typ]:
                        accum[typ][lab] = {}

                    accum[typ][lab].update(zip(labels,
                                               to_type(match["params"].split(),
                                                       float)))

        # Globals
        elif match := REs.PAIR_POT_RES["three_body_val"].match(line):
            tag, typ, lab = match.group("tag", "type", "label")

            if tag:
                typ = f"{tag}_{typ}"
            if typ not in accum:
                accum[typ] = {}

            accum[typ][lab] = to_type(match["params"], float)

    return accum


def _process_geom_table(block: Block) -> GeomTable:

    accum: GeomTable = {}
    for line in block:
        if match := REs.GEOMOPT_MIN_TABLE_RE.match(line):
            val = match.groupdict()
            fix_data_types(val, dict.fromkeys(("lambda", "fdelta", "enthalpy"), float))

            key = normalise_string(val.pop("step"))
            accum[key] = cast(dict[str, Union[bool, float]], val)

        elif match := REs.GEOMOPT_TABLE_RE.match(line):
            val = match.groupdict()
            fix_data_types(val, dict.fromkeys(("value", "tolerance"), float))

            val["converged"] = val["converged"] == "Yes"

            key = normalise_key(val.pop("parameter"))
            accum[key] = cast(dict[str, Union[bool, float]], val)

    return accum


def _process_final_config_block(block_in: Block) -> FinalConfig:

    accum: dict[str, Any] = {}
    for line in block_in:
        if block := Block.from_re(line, block_in, r"\s*Unit Cell\s*", REs.EMPTY, n_end=3):
            accum["cell"] = _process_unit_cell(block)

        elif block := Block.from_re(line, block_in,
                                    gen_table_re("Cell Contents"),
                                    gen_table_re("", "x+"), n_end=2):
            accum["atoms"] = _process_atreg_block(block)

        elif match := re.match(rf"^\s*(?:{REs.MINIMISERS_RE}):"
                               r"(?P<key>[^=]+)=\s*"
                               f"(?P<value>{REs.EXPFNUMBER_RE}).*",
                               line, re.IGNORECASE):

            key, val = normalise_key(match["key"]), to_type(match["value"], float)
            accum[key] = val

    return accum


def _process_elastic_properties(block: Block) -> ElasticProperties:
    accum: dict[str, float | ThreeVector | SixVector | ThreeByThreeMatrix] = {}
    val: float | ThreeVector | SixVector | ThreeByThreeMatrix | tuple[float, ...]

    for line in block:
        if "::" in line:
            key = line.split("::")[0]
            val = cast(Union[ThreeVector, SixVector], to_type(get_numbers(line), float))

            if len(val) == 1:
                val = val[0]

            accum[normalise_key(key)] = val
        elif blk := Block.from_re(line, block, "Speed of Sound", REs.EMPTY):

            accum["speed_of_sound"] = cast(ThreeByThreeMatrix,
                                           tuple(to_type(numbers, float)
                                                 for blk_line in blk
                                                 if (numbers := get_numbers(blk_line))),
                                           )

    return accum


def _process_internal_constraints(block: TextIO) -> list[InternalConstraints]:

    # Skip table headers
    for _ in zip(range(3), block):
        pass

    accum = []
    for line in block:
        if not line.strip():
            continue

        _, type_, target, current, *definitions = line.split()
        assert type_ in {"Bond", "Angle", "Torsion"}

        satisfied = current == "Satisfied"
        if satisfied:
            # Castep dumps current in target spot for reasons..?
            # index type current "Satisfied" ...
            target, current = None, float(target)
        else:
            # index type target current ...
            target, current = float(target), float(current)

        definitions = " ".join(definitions).split(")")
        const = {val[0]: tuple(val[1:])
                 for definition in definitions
                 if definition and
                 (val := to_type(get_numbers(definition), int))}

        elem: InternalConstraints = {"type": type_,
                                     "target": target,
                                     "current": current,
                                     "satisfied": satisfied,
                                     "constraints": const}
        accum.append(elem)

    return accum


def _process_deloc_table(block: TextIO) -> DelocInternalsTable:

    accum: DelocInternalsTable = {"constraint_mapping": {}}
    for line in block:
        if line.startswith(" constraint"):
            key, val = to_type(get_numbers(line), int)
            accum["constraint_mapping"][key] = val
        elif line.startswith("Total number of primitive"):
            type_, val = line.split()[4:6]
            type_ = "num_" + type_.strip(":")
            val = int(val)

            accum[type_] = val

    return accum


def _process_deloc_act_space_table(block: TextIO) -> DelocActiveSpace:

    accum = {}
    for line in block:
        if "active space" in line:
            accum["active_space_size"] = to_type(get_numbers(line)[0], int)
        if "degrees" in line:
            accum["num_dof"] = to_type(get_numbers(line)[0], int)
        if "primitive" in line:
            accum["num_primitive_internals"] = to_type(get_numbers(line)[0], int)

    return accum


def _process_berry_phase(block: Block) -> dict[str, Any]:
    accum = {}

    for line in block:
        if "Ionic contribution to polarisation" in line:
            table_blk = Block.from_re(line, block, REs.POL_HEADER_RE, "=+", n_end=2)
            match = REs.POL_HEADER_RE.search(line)
            assert match
            key = normalise_key(match["key"])
            accum[key] = {"units": match["unit"]}

            table_blk.remove_bounds(fore=3, back=1)  # Remove title, header and sep lines

            accum[key]["total"] = to_type(get_numbers(table_blk[-1])[2:], float)
            accum[key]["ions"] = {}
            for line in table_blk[:-1]:
                ni, nsp, _z_ion, a, b, c, *pol = line.split()
                accum[key]["ions"][to_type((nsp, ni), int)] = {"coords": to_type((a, b, c), float),
                                                               "polarisation": to_type(pol, float)}

        elif match := REs.POL_TABLE_RE.search(line):
            key = normalise_key(match["key"])
            accum[key] = {"units": match["unit"],
                          "val": to_type(match.group("a", "b", "c"), float)}

        elif table_blk := Block.from_re(line, block, REs.POL_HEADER_RE, "=+", n_end=2,
                                        eof_possible=True):
            match = REs.POL_HEADER_RE.search(line)
            assert match
            key = normalise_key(match["key"])

            table_blk.remove_bounds(fore=2, back=1)

            if key == "polarisation" and "Quantum" not in line:  # Ordinary polarisation
                accum.setdefault(key, {})
                if match["unit"] != "phase":
                    accum[key]["units"] = match["unit"]
                    subkey = "val"
                else:
                    subkey = "phase"

                accum[key].setdefault(subkey, {})

                for line in table_blk:
                    curr = line.split()[0]
                    accum[key][subkey][curr] = to_type(get_numbers(line), float)

            elif key == "polarisation":
                accum.setdefault(key, {})

                # Extra line of description.

                table_blk.remove_bounds(fore=1, back=0)
                for line in table_blk:
                    _, *direc, elec, ionic, total, quantum = to_type(get_numbers(line), float)
                    accum[key][tuple(direc)] = {"elec": elec,
                                                "ionic": ionic,
                                                "total": total,
                                                "quantum": quantum,
                                                "units": match["unit"]}

            elif key == "volume_polarisation":

                accum[key] = {"units": match["unit"]}

                for line in table_blk:
                    curr = line.split()[0]
                    accum[key][curr] = to_type(get_numbers(line), float)

    return accum



"""Parse castep .elastic files."""
from __future__ import annotations

from collections import defaultdict
from typing import TextIO

from ..utilities.castep_res import ELASTIC_BLOCK_RE, ELASTIC_INTERNAL_STRAIN_RE, get_numbers
from ..utilities.filewrapper import Block
from ..utilities.utility import atreg_to_index, file_or_path, normalise_key, to_type
from .parse_utilities import parse_regular_header


@file_or_path(mode="r")
def parse_elastic_file(elastic_file: TextIO) -> dict[str, list[list[float]]]:
    """
    Parse castep .elastic files.

    Parameters
    ----------
    elastic_file
        Open handle to file to parse.

    Returns
    -------
    dict[str, list[list[float]]]
        Parsed info.
    """
    accum = {}

    for line in elastic_file:
        if block := Block.from_re(line, elastic_file,
                                  "BEGIN header", "END header"):
            accum.update(parse_regular_header(block))
            accum.pop("")

        elif block := Block.from_re(line, elastic_file,
                                    "BEGIN Internal Strain", "END"):
            match = ELASTIC_BLOCK_RE.match(next(block))
            key = normalise_key(match["key"]).removesuffix("_xx_yy_zz_yz_zx_xy")
            accum[key] = defaultdict(list)
            accum[key]["units"] = match["unit"]

            block.remove_bounds(fore=0, back=1)
            for blk_line in block:
                match = ELASTIC_INTERNAL_STRAIN_RE.match(blk_line).groupdict()
                atom = atreg_to_index(match)
                accum[key][atom].append(to_type(get_numbers(blk_line), float))

        elif block := Block.from_re(line, elastic_file,
                                    "BEGIN", "END"):
            match = ELASTIC_BLOCK_RE.match(next(block))
            key = normalise_key(match["key"]).removesuffix("_xx_yy_zz_yz_zx_xy")

            accum[key] = {"units": match["unit"]}
            accum[key]["val"] = tuple(to_type(numbers, float)
                                      for blk_line in block
                                      if (numbers := get_numbers(blk_line)))

    return accum



"""Test reading binary files."""
from __future__ import annotations

from pathlib import Path

import pytest
from dump_fortran_unformatted import data_types, fake_file, raw_data, to_bytes

from castep_outputs.bin_parsers.fortran_bin_parser import binary_file_reader
from castep_outputs.utilities.utility import to_type

DATA_FILES = Path(__file__).parent / "data_files"
DEFAULT_SAMPLE = (1, 2, 3, 2, 3.1, "Hello", (1., 3., 6.))


@pytest.mark.parametrize("fake_file, data_types, raw_data", [
    (DEFAULT_SAMPLE,)*3,
], indirect=True)
def test_binary_file_reader(fake_file, data_types, raw_data):
    """Test reading a generic "file"."""
    reader = binary_file_reader(fake_file)

    for datum, typ, expected in zip(reader, data_types, raw_data):
        assert to_type(datum, typ) == expected


def test_actual_read():
    """Test reading a castep `.cst_esp` file elements."""
    dtypes = ((int, 1), (int, (16, 16, 16)))

    with (DATA_FILES / "test.cst_esp").open("rb") as file:
        reader = binary_file_reader(file)
        for (typ, expected), datum in zip(dtypes, reader):
            assert to_type(datum, typ) == expected

        for (n, datum) in enumerate(reader):
            x, y = divmod(n, 16)
            ind = to_type(datum[:8], int)
            assert ind == (x + 1, y + 1)  # 1 indexed
            res = to_type(datum[8:], complex)
            assert len(res) == dtypes[1][1][2]

@pytest.mark.parametrize("fake_file, raw_data", [
    (DEFAULT_SAMPLE,)*2,
], indirect=True)
@pytest.mark.parametrize("forward, skip, index", [
    (2, True, 1),
    (1, 3, 4),
    (3, -2, 1),
    (2, -1, 1),
    (2, 0, 2),
])
def test_rewind(fake_file, raw_data, forward, skip, index):
    """Test rewind functionality of raw_file_reader."""
    reader = binary_file_reader(fake_file)

    for i, val in zip(range(forward), reader):
        assert to_bytes(raw_data[i]) == val

    result = reader.send(skip)
    assert to_bytes(raw_data[index]) == result


if __name__ == "__main__":
    pytest.main()



"""List of parsers for file formats."""
from __future__ import annotations

from collections.abc import Callable

from .bands_file_parser import parse_bands_file
from .castep_file_parser import parse_castep_file
from .cell_param_file_parser import parse_cell_param_file
from .efield_file_parser import parse_efield_file
from .elastic_file_parser import parse_elastic_file
from .epme_file_parser import parse_epme_file
from .err_file_parser import parse_err_file
from .hug_file_parser import parse_hug_file
from .magres_file_parser import parse_magres_file
from .md_geom_file_parser import parse_md_geom_file
from .parse_fmt_files import (
                              parse_chdiff_fmt_file,
                              parse_den_fmt_file,
                              parse_elf_fmt_file,
                              parse_pot_fmt_file,
)
from .phonon_dos_file_parser import parse_phonon_dos_file
from .phonon_file_parser import parse_phonon_file
from .tddft_file_parser import parse_tddft_file
from .ts_file_parser import parse_ts_file
from .xrd_sf_file_parser import parse_xrd_sf_file

__all__ = [
    "parse_bands_file",
    "parse_castep_file",
    "parse_cell_param_file",
    "parse_cell_param_file",
    "parse_chdiff_fmt_file",
    "parse_den_fmt_file",
    "parse_efield_file",
    "parse_elastic_file",
    "parse_elf_fmt_file",
    "parse_epme_file",
    "parse_err_file",
    "parse_hug_file",
    "parse_magres_file",
    "parse_md_geom_file",
    "parse_md_geom_file",
    "parse_phonon_dos_file",
    "parse_phonon_file",
    "parse_pot_fmt_file",
    "parse_tddft_file",
    "parse_ts_file",
    "parse_xrd_sf_file",
]


#: Dictionary of available parsers.
PARSERS: dict[str, Callable] = {
    "castep": parse_castep_file,
    "cell": parse_cell_param_file,
    "param": parse_cell_param_file,
    "geom": parse_md_geom_file,
    "md": parse_md_geom_file,
    "bands": parse_bands_file,
    "hug": parse_hug_file,
    "phonon_dos": parse_phonon_dos_file,
    "efield": parse_efield_file,
    "xrd_sf": parse_xrd_sf_file,
    "elf_fmt": parse_elf_fmt_file,
    "chdiff_fmt": parse_chdiff_fmt_file,
    "pot_fmt": parse_pot_fmt_file,
    "den_fmt": parse_den_fmt_file,
    "elastic": parse_elastic_file,
    "ts": parse_ts_file,
    "magres": parse_magres_file,
    "tddft": parse_tddft_file,
    "err": parse_err_file,
    "phonon": parse_phonon_file,
    "epme": parse_epme_file,
    }

#: Names of parsers/parsable file extensions (without ``"."``).
CASTEP_OUTPUT_NAMES: tuple[str, ...] = tuple(PARSERS.keys())

#: Names of parsable file extensions.
CASTEP_FILE_FORMATS: tuple[str, ...] = tuple(f".{typ}" for typ in CASTEP_OUTPUT_NAMES)



"""Module containing argument parser and processing."""
from __future__ import annotations

import argparse
from collections.abc import Sequence
from pathlib import Path

from castep_outputs import __version__
from castep_outputs.bin_parsers import CASTEP_FILE_FORMATS as BIN_FORMATS
from castep_outputs.bin_parsers import CASTEP_OUTPUT_NAMES as BIN_NAMES
from castep_outputs.parsers import CASTEP_FILE_FORMATS, CASTEP_OUTPUT_NAMES
from castep_outputs.utilities.dumpers import SUPPORTED_FORMATS

# pylint: disable=line-too-long
ALL_FORMATS = (*CASTEP_FILE_FORMATS, *BIN_FORMATS)
ALL_NAMES = (*CASTEP_OUTPUT_NAMES, *BIN_NAMES)

#: Main argument parser for castep outputs.
ARG_PARSER = argparse.ArgumentParser(
    prog="castep_outputs",
    description=f"""Attempts to find all files for seedname, filtered by `inc` args (default: all).
    Explicit files can be passed using longname arguments.
    castep_outputs can parse most castep outputs including: {', '.join(ALL_FORMATS)}""",
)

ARG_PARSER.add_argument("seedname", nargs=argparse.REMAINDER, help="Seed name for data")
ARG_PARSER.add_argument("-V", "--version", action="version", version=f"%(prog)s v{__version__}")
ARG_PARSER.add_argument("-L", "--log", help="Verbose output",
                        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
                        default="WARNING")
ARG_PARSER.add_argument("-o", "--output", help="File to write output, default: screen",
                        default=None)
ARG_PARSER.add_argument("-f", "--out-format", help="Output format", choices=SUPPORTED_FORMATS,
                        default="json")

ARG_PARSER.add_argument("-t", "--testing", action="store_true",
                        help="Set testing mode to produce flat outputs")

ARG_PARSER.add_argument("-A", "--inc-all", action="store_true",
                        help="Extract all available information")

for output_name in ALL_NAMES:
    ARG_PARSER.add_argument(f"--inc-{output_name}", action="store_true",
                            help=f"Extract .{output_name} information")

for output_name in ALL_NAMES:
    ARG_PARSER.add_argument(f"--{output_name}", nargs="*",
                            help=f"Extract from {output_name.upper()} as .{output_name} type",
                            default=[])


def parse_args(to_parse: Sequence[str] = ()) -> argparse.Namespace:
    """
    Parse all arguments and add those caught by flags.

    Parameters
    ----------
    to_parse
        Arguments to handle in this call.

    Returns
    -------
    argparse.Namespace
        Parsed args.

    Examples
    --------
    parse_args()
    """
    args = ARG_PARSER.parse_args()

    parse_all = args.inc_all or not any(getattr(args, f"inc_{typ}") for typ in ALL_NAMES)

    # Set all flags
    if parse_all and not to_parse:
        for typ in ALL_NAMES:
            setattr(args, f"inc_{typ}", True)

    # Only parse those which are requested
    for typ in to_parse:
        setattr(args, f"inc_{typ}", True)

    # Add seeded files into parse list if to be included
    for seed in args.seedname:
        seed = Path(seed)
        if seed.is_file() and (ext := seed.suffix[1:]) in ALL_NAMES:
            getattr(args, ext).append(str(seed))
        else:
            for typ in ALL_NAMES:
                trial = seed.with_suffix(f".{typ}")
                if getattr(args, f"inc_{typ}") and trial.is_file():
                    getattr(args, typ).append(str(trial))

            if args.inc_err and (err_files := seed.parent.glob(f"{seed}.*.err")):
                args.err.extend(map(str, err_files))

    return args


def extract_parsables(args: argparse.Namespace) -> dict[str, list[str]]:
    """
    Extract the files to parse from the ``Namespace``.

    Parameters
    ----------
    args
        Namespace to process.

    Returns
    -------
    dict[str, list[str]]
        Files to parse.
    """
    return {typ: getattr(args, typ) for typ in ALL_NAMES}



"""
Attempt to guess the files which will be output by the castep calculation.

Notes
-----
This is based on the working version of CASTEP, so many not exactly align with
the last release.
"""

from __future__ import annotations

import re
from copy import deepcopy
from enum import Enum, auto
from functools import singledispatch
from itertools import combinations_with_replacement
from pathlib import Path
from typing import TypeVar

from ..parsers.cell_param_file_parser import CellParamData, parse_cell_param_file

Self = TypeVar("Self", bound="UCEnum")


class UCEnum(Enum):
    """Auto upperclass enum."""

    @classmethod
    def _missing_(cls, task: str) -> Self:
        task = task.upper()
        return cls[task]


class Task(UCEnum):
    """CASTEP tasks."""

    SINGLEPOINT = auto()
    PHONON = auto()
    PHONON_EFIELD = auto()
    EFIELD = auto()
    THERMODYNAMICS = auto()
    EPCOUPLING = auto()
    # Not implemented yet.
    # ELASTIC = auto()
    GEOMOPT = auto()
    MD = auto()
    SOCKET_DRIVER = auto()
    TRANSITION_STATE_SEARCH = auto()
    WANNIER = auto()
    MAGRES = auto()
    SPECTRAL = auto()
    AUTOSOLVATION = auto()

    # Aliases
    GEOMETRYOPTIMISATION = GEOMOPT
    GEOMETRYOPTIMIZATION = GEOMOPT
    GEOM = GEOMOPT

    SPE = SINGLEPOINT
    ENERGY = SINGLEPOINT

    MOLECULARDYNAMICS = MD
    DYNAMICS = MD

    SOCKETDRIVER = SOCKET_DRIVER

    TSS = TRANSITION_STATE_SEARCH
    TRANSITIONSTATESEARCH = TRANSITION_STATE_SEARCH

    ELECTRONICSPECTROSCOPY = SPECTRAL
    BANDSTRUCTURE = SPECTRAL
    ELNES = SPECTRAL
    OPTICS = SPECTRAL


class MagresTask(UCEnum):
    """Magres tasks."""

    SHIELDING = auto()
    EFG = auto()
    NMR = auto()
    HYPERFINE = auto()
    GTENSOR = auto()
    EPR = auto()
    JCOUPLING = auto()


class SpectralTask(UCEnum):
    """Spectral tasks."""

    CORELOSS = auto()
    BANDSTRUCTURE = auto()
    OPTICS = auto()
    DOS = auto()
    ALL = auto()


class SpectralTheory(UCEnum):
    """Spectral theories."""

    TDDFT = auto()
    DFT = auto()


XC_TO_PS_THEORY = {
    "pw91": "pbe",
    "pbe": "pbe",
    "rpbe": "rpbe",
    "pbesol": "pbesol",
    "wc": "wc",
    "blyp": "blyp",
    "ms2": "ms2",
    "rscan": "rscan",
    "r2scan": "rscan",
    "lda_pw": "lda_pw",
}


def get_spectral_files(
    param_data: CellParamData,
    seedname: str,
    *,
    is_nlxc: bool,
) -> set[str]:
    """Get files associated with Task = Spectral.

    Parameters
    ----------
    param_data : CellParamData
        Dict containing parameters.
    seedname : str
        Current seedname.
    is_nlxc : bool
        Whether XC functional is non-local.

    Returns
    -------
    set[str]
        Files generated in spectral task.

    Raises
    ------
    KeyError
        Impossible combination.
    """
    out_files = set()
    devel_code = param_data.get("devel_code", {})
    raw_task = param_data.get("task", "SINGLEPOINT").upper()
    spectral_theory = SpectralTheory(param_data.get("spectral_theory", "dft"))
    spectral_task = SpectralTask(param_data.get("spectral_task", "bandstructure"))

    out_files.add(f"{seedname}-out.cell")

    if param_data.get("spectral_restart", True):
        out_files.add(f"{seedname}.*.spec")

    if param_data.get("write_orbitals") or raw_task == "BANDSTRUCTURE":
        out_files.add(f"{seedname}.orbitals")
    if spectral_theory is SpectralTheory.TDDFT:
        if spectral_task is SpectralTask.OPTICS:
            spec_calc = {"core": False, "ome": False, "dome": True}
        elif spectral_task in {SpectralTask.DOS, SpectralTask.BANDSTRUCTURE}:
            spec_calc = {"core": False, "ome": False, "dome": False}
        elif spectral_task is SpectralTask.ALL:
            spec_calc = {"core": False, "ome": False, "dome": True}
        else:
            raise KeyError("Invalid param file")
    else:  # noqa: PLR5501
        if spectral_task is SpectralTask.CORELOSS:
            if is_nlxc:
                raise KeyError("Invalid param file")
            spec_calc = {"core": True, "ome": True, "dome": False}
        elif spectral_task is SpectralTask.OPTICS:
            spec_calc = {"core": False, "ome": False, "dome": True}
        elif spectral_task is SpectralTask.DOS:
            spec_calc = {"core": False, "ome": not is_nlxc, "dome": False}
        elif spectral_task is SpectralTask.BANDSTRUCTURE:
            spec_calc = {"core": False, "ome": False, "dome": False}
        elif spectral_task is SpectralTask.ALL:
            spec_calc = {"core": not is_nlxc, "ome": not is_nlxc, "dome": not is_nlxc}
        else:
            raise KeyError("Invalid param file")

    spectral_devel = devel_code.get("spectral", {})
    spec_calc["core"] = spectral_devel.get("calc_core", spec_calc["core"])
    spec_calc["ome"] = spectral_devel.get("calc_ome", spec_calc["ome"])
    spec_calc["pdos"] = spectral_devel.get("calc_pdos", param_data.get("pdos_calculate_weights"))
    spec_calc["dome"] = spectral_devel.get("calc_dome", spec_calc["dome"])

    out_files |= {
        f"{seedname}.{curr}_bin"
        if spectral_theory is SpectralTheory.DFT
        else f"{seedname}_tddft.{curr}_bin"
        for curr in ("core", "ome", "pdos", "dome")
        if spec_calc[curr]
    }

    return out_files


def get_xc_info(param_data: CellParamData) -> set[str]:
    """Get relevant info from the XC parameters.

    Returns a reduced form of libxc

    Parameters
    ----------
    param_data : CellParamData
        Param file data containing functional definition.

    Returns
    -------
    set[str]
        Active xc functionals.

    Examples
    --------
    >>> get_xc_info({"xc_functional": "pbe"})
    {'pbe'}
    >>> get_xc_info({"xc_functional": "pbe",  # xc_definition takes priority like castep.
    ...              "xc_definition": {"xc": {"lda": 1.}}})
    {'lda'}
    >>> sorted(   # Sorted to force set order.
    ... get_xc_info({"xc_definition": {"xc": {"pbe": 0.25,
    ...                                       "libxc_gga_x_2d_b86_mgc": 0.25,
    ...                                       "libxc_lda_c_vwn": 0.25,
    ...                                       "libxc_hyb_mgga_xc_revtpssh": 0.25}}})
    ... )
    ['libxc_gga', 'libxc_hyb_mgga', 'libxc_lda', 'pbe']
    """
    xc_f = param_data.get("xc_functional", "lda").lower()
    xc_d = param_data.get("xc_definition")

    xc = xc_d["xc"].keys() if xc_d else {xc_f}
    xc = map(str.lower, xc)

    return {typ[0] if (typ := re.match(r"libxc(_hyb)?_(m?gga|lda)", key, re.IGNORECASE)) else key
            for key in xc}


@singledispatch
def get_generated_files(
    seedname: Path | str = "seedname",
    /,
    *,
    param_file: Path | None = None,
    cell_file: Path | None = None,
) -> list[str]:
    """Predict files which would be produced by running inputs.

    Parameters
    ----------
    seedname : Path | str, optional
        Seedname of files to parse.
    param_file : Path | str, optional
        Pre-parsed param file.
    cell_file : Path | str, optional
        Pre-parsed cell file.

    Returns
    -------
    set[str]
        Files which would be produced.

    Notes
    -----
    May struggle to get exact information from complex xc_definitions.
    """
    seedname = Path(seedname)
    if param_file is None:
        param_file = seedname.with_suffix(".param")
    if cell_file is None:
        cell_file = seedname.with_suffix(".cell")

    param = parse_cell_param_file(param_file)[0] if param_file.exists() else {}
    cell = parse_cell_param_file(cell_file)[0] if cell_file.exists() else {}
    return get_generated_files(param, cell, seedname.with_suffix(""))


@get_generated_files.register(dict)
def _(
    param_data: CellParamData,
    cell_data: CellParamData | None = None,
    src: Path | str = "seedname",
) -> list[str]:
    if cell_data is None:
        cell_data = {}

    seedname = Path(src).stem
    param_data = deepcopy(param_data)
    devel_code = param_data.get("devel_code", {})
    raw_task = param_data.get("task", "SINGLEPOINT").upper()
    task = Task(raw_task)

    xc = get_xc_info(param_data)

    is_mgga = xc & {"ms2", "rscan", "r2scan", "libxc_mgga", "libxc_hyb_mgga"}
    is_nlxc = bool(xc & {
        "hf",
        "hf-lda",
        "shf",
        "sx-lda",
        "shf-lda",
        "pbe0",
        "b3lyp",
        "hse03",
        "hse06",
        "spbe0",
    })
    is_oep = xc & {"oep", "lfx", "elp", "ceda"}
    is_spin = (
        param_data.get("spin_treatment", "NONE").upper() == "VECTOR"
        or param_data.get("spin_orbit_coupling")
        or param_data.get("spin_polarised")
        or param_data.get("spin_polarized")
    )
    is_pp = devel_code.get("_pp", False)
    tddft_on = param_data.get("tddft_selected_state", 0) > 0

    out_files = {f"{seedname}.castep"}
    if param_data.get("write_none"):
        for i in (
            "write_bib",
            "write_checkpoint",
            "write_cst_esp",
            "write_bands",
            "write_geom",
            "write_md",
        ):
            param_data[i] = False

    write_check = param_data.get("write_checkpoint", "ALL").upper()
    if write_check in {True, "TRUE", "ALL", "BOTH", "FULL"}:
        out_files.add(f"{seedname}.check")
        out_files.add(f"{seedname}.castep_bin")
    elif write_check == "MINIMAL":
        out_files.add(f"{seedname}.castep_bin")
    elif write_check in {False, "FALSE", "NONE"}:
        pass
    else:
        raise NotImplementedError(
            f"Cannot understand checkpoint: {param_data['write_checkpoint']!r}",
        )

    if param_data.get("write_bands", True) and not is_pp:
        out_files.add(
            f"{seedname}.bands"
            if not param_data.get("tddft_selected_state", 0)
            else f"{seedname}_tddft.bands",
        )

    if devel_code.get("write_formatted_bands"):
        out_files.add(f"{seedname}_*.orbit_fmt")

    if param_data.get("write_bib", True):
        out_files.add(f"{seedname}.bib")
    if not is_pp and param_data.get("write_cst_esp", True):
        out_files.add(f"{seedname}.cst_esp")

    if param_data.get("write_formatted_potential"):
        out_files.add(f"{seedname}.pot_fmt")
        if is_oep:
            out_files.add(f"{seedname}.oep_fmt")

    if param_data.get("write_formatted_density"):
        out_files.add(f"{seedname}.den_fmt")

    if param_data.get("calculate_densdiff"):
        out_files.add(f"{seedname}.chdiff")
        if param_data.get("write_formatted_density"):
            out_files.add(f"{seedname}.chdiff_fmt")

    if param_data.get("calculate_elf"):
        out_files.add(f"{seedname}.elf")
        # Commented out
        # out_files.add("rho_test.den_fmt")
        if param_data.get("write_formatted_elf"):
            out_files.add(f"{seedname}.elf_fmt")

    if param_data.get("write_formatted_dielec_perm"):
        out_files.add(f"{seedname}.eps_format")

    if param_data.get("write_orbitals"):
        pass
    if param_data.get("write_cif_structure"):
        out_files.add(f"{seedname}-out.cif")
    if param_data.get("write_cell_structure"):
        out_files.add(f"{seedname}-out.cell")

    if param_data.get("efield_calc_ion_permittivity"):
        out_files.add(f"{seedname}.efield")

    if param_data.get("calculate_modos"):
        out_files.add(f"{seedname}.modos_state*")

    # Devel enabled
    if devel_code.get("zdos_scf"):
        out_files.add(f"{seedname}.scf_zdos")
    if devel_code.get("write_unformatted_density"):
        out_files.add(f"{seedname}.altden")
    if devel_code.get("trace") or devel_code.get("prof") or devel_code.get("profclass"):
        out_files.add(f"{seedname}.*.profile")
    if devel_code.get("write_formatted_bands"):
        out_files.add(f"{seedname}_*.orbit_fmt")
    if devel_code.get("pp_hybrid") and task in {Task.GEOMOPT, Task.MD}:
        out_files.add(f"{seedname}.hybrid-md.xyz")
    if is_spin and devel_code.get("lda_sf_pot_write"):
        out_files.add(f"{seedname}.B_xc.pot_fmt")
    if devel_code.get("calculate_xrd_sf"):
        out_files.add(f"{seedname}.xrd_sf")
    if devel_code.get("zdos_spectral") and task is Task.SPECTRAL:
        out_files.add(f"{seedname}.spectral_zdos")

    if devel_code.get("_pp") and (pp := devel_code.get("pp")):
        if pp.get("pot_print"):
            if cell_data:
                spec = {spec: None for spec, _ind in cell_data.get("positions_frac", {})}
                out_files |= {
                    f"{seedname}_{a}_{b}.pot" for a, b in combinations_with_replacement(spec, 2)
                }

            else:
                out_files.add(f"{seedname}_*_.pot")

        if pp.get("fd_check"):
            out_files.add(f"{seedname}.fd")

    if param_data.get("pdos_calculate_weights") and raw_task not in {
        "BANDSTRUCTURE",
        "ELECTRONICSPECTROSCOPY",
    }:
        out_files.add(f"{seedname}.pdos_weights")

    is_usp = True
    for spec, pot in cell_data.get("species_pot", {}).items():
        if isinstance(pot, str):
            is_usp = is_usp or pot.endswith(".usp", ".uspcc", ".uspso")
            continue

        if xc & {"libxc_mgga", "libxc_hyb_mgga"}:
            theory = "RSCAN"
        elif xc & {"libxc_gga", "libxc_hyb_gga"}:
            theory = "PBE"
        else:
            theory = XC_TO_PS_THEORY.get(next(iter(xc)), "LDA")

        otf_name = f"{spec}_EXT_{theory}_OTF"

        if pot.get("print"):
            out_files.add(f"{otf_name}.gamma")
            out_files.add(f"{otf_name}.pwave")
            out_files.add(f"{otf_name}.beta")
            out_files.add(f"{otf_name}.econv")

        # Commented out 25.1
        # out_files.add(f"{otf_name}.orbs")
        # out_files.add(f"{otf_name}.potential")

        if not is_mgga and param_data.get("write_otfg", True):
            if is_usp:
                if is_spin:
                    out_files.add(f"{otf_name}.uspso")
                else:
                    out_files.add(f"{otf_name}.usp")
            else:
                out_files.add(f"{otf_name}.recpot")

    if devel_code.get("calc_elastic") or "elastic" in devel_code:
        out_files.add(f"{seedname}.elastic")
        if (
            devel_code.get("elastic", {}).get("deform_pot")
            and not is_mgga
            and not is_usp
            and not is_spin
            and not is_nlxc
        ):
            out_files.add(f"{seedname}.dbands_dstrain")

    if task is Task.AUTOSOLVATION:
        out_files.add(f"{seedname}.vacuum_den")
    # elif task is Task.ELASTIC:  # Not implemented yet
    #     out_files.add(f"{seedname}.elastic")
    #     if (
    #             devel_code.get("ELASTIC", {}).get("DEFORM_POT") and
    #             not is_mgga and
    #             not is_usp and
    #             not is_spin and
    #             not is_nlxc
    #     ):
    #         out_files.add(f"{seedname}.dbands_dstrain")

    elif task is Task.EPCOUPLING:
        out_files.add(f"{seedname}.epme")
        out_files.add(f"{seedname}-*-*-*.ep_pot")
    elif task is Task.GEOMOPT:
        if param_data.get("write_geom", True):
            out_files.add(f"{seedname}.geom")
        if param_data.get("geom_method") == "pes":
            out_files.add(f"{seedname}.pes")
    elif task is Task.SPECTRAL:
        out_files |= get_spectral_files(
            param_data,
            seedname,
            is_nlxc=is_nlxc,
        )
    elif task is Task.MD:
        if param_data.get("write_md", True):
            out_files.add(f"{seedname}.md")
        if param_data.get("md_ensemble") in {"nvhug", "nphug"}:
            out_files.add(f"{seedname}.hug")

        extrap = param_data.get("md_extrap", "first").lower()
        opt_mem = (
            param_data.get("opt_strategy", "default").lower() == "memory"
            or param_data.get("opt_strategy_bias", 0) >= 0
        )
        mix_method = param_data.get("metals_method", "dm").lower()
        if mix_method == "dm" and extrap in {"first", "second", "mixed"} and opt_mem:
            out_files.add(f"{seedname}.*.wfm")
            out_files.add(f"{seedname}.*.drhom")
        if mix_method == "dm" and extrap in {"second", "mixed"} and opt_mem:
            out_files.add(f"{seedname}.*.wf2m")
            out_files.add(f"{seedname}.*.drho2m")

        if param_data.get("md_num_beads", 1) > 1 and param_data.get("num_farms", 1) > 1:
            out_files |= {
                f"{seedname}_farm{i:0>3d}.castep" for i in range(param_data.get("num_farms"))
            }

    elif task is Task.MAGRES:
        out_files.add(f"{seedname}.magres")
        if param_data.get("magres_write_response"):
            out_files.add("current.dat")
            out_files.add("gsden.cube")
    elif task is Task.TSS:
        out_files.add(f"{seedname}.ts")
    elif task is Task.THERMODYNAMICS:
        ...  # No extra files? Created in secondd?

    if task in {Task.PHONON, Task.PHONON_EFIELD}:
        out_files.add(f"{seedname}.phonon")
        if param_data.get("phonon_calculate_dos"):
            out_files.add(f"{seedname}.phonon_dos")
        if filename := devel_code.get("phonon", {}).get("write_external_born"):
            out_files.add(filename)
        if devel_code.get("phonon", {}).get("write_first_order_potential"):
            out_files.add(f"{seedname}-*-*-*.ep_pot")

    if task in {Task.EFIELD, Task.PHONON_EFIELD}:
        out_files.add(f"{seedname}.efield")

    if task is Task.WANNIER:
        raise KeyError("Wannier is not supported")
    if task is Task.SOCKET_DRIVER:
        raise NotImplementedError("Unable to know what files may be written as socket driver.")

    if tddft_on:
        out_files.add(f"{seedname}.tddft")
        if devel_code.get("tddft", {}).get("den1_write"):
            out_files.add(f"{seedname}.td_den")
        if devel_code.get("tddft", {}).get("save_td_den_fmt"):
            out_files.add("den_dump*.dat")

    return sorted(out_files)
